"""Tests for schematic model classes to increase coverage.

Tests for:
- Wire, Junction, Label, HierarchicalLabel, PowerSymbol (elements.py)
- SymbolDef, SymbolInstance (symbol.py)
- Schematic class (schematic.py)
"""

import pytest

from kicad_tools.schematic.models.elements import (
    HierarchicalLabel,
    Junction,
    Label,
    PowerSymbol,
    Wire,
)
from kicad_tools.schematic.models.pin import Pin
from kicad_tools.schematic.models.schematic import Schematic, SnapMode
from kicad_tools.schematic.models.symbol import SymbolDef, SymbolInstance
from kicad_tools.sexp import SExp


class TestWire:
    """Tests for Wire dataclass."""

    def test_wire_creation(self):
        """Create wire with explicit coordinates."""
        wire = Wire(x1=10.0, y1=20.0, x2=30.0, y2=40.0)
        assert wire.x1 == 10.0
        assert wire.y1 == 20.0
        assert wire.x2 == 30.0
        assert wire.y2 == 40.0
        assert wire.uuid_str is not None

    def test_wire_between(self):
        """Create wire between two points."""
        wire = Wire.between((10.5, 20.5), (30.5, 40.5))
        assert wire.x1 == 10.5
        assert wire.y1 == 20.5
        assert wire.x2 == 30.5
        assert wire.y2 == 40.5

    def test_wire_between_rounds_coordinates(self):
        """Wire.between rounds to 2 decimal places."""
        wire = Wire.between((10.123456, 20.789), (30.001, 40.999))
        assert wire.x1 == 10.12
        assert wire.y1 == 20.79
        assert wire.x2 == 30.0
        assert wire.y2 == 41.0

    def test_wire_to_sexp_node(self):
        """Wire generates valid S-expression node."""
        wire = Wire(x1=10.0, y1=20.0, x2=30.0, y2=40.0, uuid_str="test-uuid")
        sexp = wire.to_sexp_node()
        assert sexp.name == "wire"
        # Check it has pts node
        pts = sexp.get("pts")
        assert pts is not None

    def test_wire_to_sexp(self):
        """Wire generates S-expression string."""
        wire = Wire(x1=10.0, y1=20.0, x2=30.0, y2=40.0)
        sexp_str = wire.to_sexp()
        assert "wire" in sexp_str
        assert "pts" in sexp_str

    def test_wire_from_sexp(self):
        """Parse wire from S-expression node."""
        # Build a wire sexp manually
        wire_sexp = SExp.list(
            "wire",
            SExp.list(
                "pts",
                SExp.list("xy", 10.0, 20.0),
                SExp.list("xy", 30.0, 40.0),
            ),
            SExp.list("uuid", "test-uuid-123"),
        )

        wire = Wire.from_sexp(wire_sexp)
        assert wire.x1 == 10.0
        assert wire.y1 == 20.0
        assert wire.x2 == 30.0
        assert wire.y2 == 40.0
        assert wire.uuid_str == "test-uuid-123"

    def test_wire_from_sexp_missing_points(self):
        """Wire parsing fails with insufficient points."""
        wire_sexp = SExp.list(
            "wire",
            SExp.list("pts", SExp.list("xy", 10.0, 20.0)),  # Only one point
            SExp.list("uuid", "test-uuid"),
        )

        with pytest.raises(ValueError, match="at least 2 xy points"):
            Wire.from_sexp(wire_sexp)


class TestJunction:
    """Tests for Junction dataclass."""

    def test_junction_creation(self):
        """Create junction at position."""
        junc = Junction(x=10.0, y=20.0)
        assert junc.x == 10.0
        assert junc.y == 20.0
        assert junc.uuid_str is not None

    def test_junction_rounds_coordinates(self):
        """Junction rounds coordinates on creation."""
        junc = Junction(x=10.123456, y=20.789)
        assert junc.x == 10.12
        assert junc.y == 20.79

    def test_junction_to_sexp_node(self):
        """Junction generates valid S-expression node."""
        junc = Junction(x=10.0, y=20.0, uuid_str="junc-uuid")
        sexp = junc.to_sexp_node()
        assert sexp.name == "junction"

    def test_junction_to_sexp(self):
        """Junction generates S-expression string."""
        junc = Junction(x=10.0, y=20.0)
        sexp_str = junc.to_sexp()
        assert "junction" in sexp_str

    def test_junction_from_sexp(self):
        """Parse junction from S-expression node."""
        junc_sexp = SExp.list(
            "junction",
            SExp.list("at", 10.0, 20.0),
            SExp.list("uuid", "junc-uuid-123"),
        )

        junc = Junction.from_sexp(junc_sexp)
        assert junc.x == 10.0
        assert junc.y == 20.0
        assert junc.uuid_str == "junc-uuid-123"


class TestLabel:
    """Tests for Label dataclass."""

    def test_label_creation(self):
        """Create label with text and position."""
        label = Label(text="NET1", x=10.0, y=20.0)
        assert label.text == "NET1"
        assert label.x == 10.0
        assert label.y == 20.0
        assert label.rotation == 0
        assert label.uuid_str is not None

    def test_label_with_rotation(self):
        """Create label with rotation."""
        label = Label(text="NET1", x=10.0, y=20.0, rotation=90)
        assert label.rotation == 90

    def test_label_to_sexp_node(self):
        """Label generates valid S-expression node."""
        label = Label(text="NET1", x=10.0, y=20.0, uuid_str="label-uuid")
        sexp = label.to_sexp_node()
        assert sexp.name == "label"

    def test_label_to_sexp(self):
        """Label generates S-expression string."""
        label = Label(text="NET1", x=10.0, y=20.0)
        sexp_str = label.to_sexp()
        assert "label" in sexp_str
        assert "NET1" in sexp_str

    def test_label_from_sexp(self):
        """Parse label from S-expression node."""
        label_sexp = SExp.list(
            "label",
            "NET1",
            SExp.list("at", 10.0, 20.0, 45),
            SExp.list("uuid", "label-uuid-123"),
        )

        label = Label.from_sexp(label_sexp)
        assert label.text == "NET1"
        assert label.x == 10.0
        assert label.y == 20.0
        assert label.rotation == 45
        assert label.uuid_str == "label-uuid-123"

    def test_label_from_sexp_no_rotation(self):
        """Parse label without rotation defaults to 0."""
        label_sexp = SExp.list(
            "label",
            "NET1",
            SExp.list("at", 10.0, 20.0),
            SExp.list("uuid", "label-uuid"),
        )

        label = Label.from_sexp(label_sexp)
        assert label.rotation == 0


class TestHierarchicalLabel:
    """Tests for HierarchicalLabel dataclass."""

    def test_hier_label_creation(self):
        """Create hierarchical label with defaults."""
        hl = HierarchicalLabel(text="DATA_OUT", x=10.0, y=20.0)
        assert hl.text == "DATA_OUT"
        assert hl.x == 10.0
        assert hl.y == 20.0
        assert hl.shape == "input"
        assert hl.rotation == 0

    def test_hier_label_with_shape(self):
        """Create hierarchical label with specific shape."""
        hl = HierarchicalLabel(text="DATA_OUT", x=10.0, y=20.0, shape="output", rotation=180)
        assert hl.shape == "output"
        assert hl.rotation == 180

    def test_hier_label_to_sexp_node(self):
        """HierarchicalLabel generates valid S-expression node."""
        hl = HierarchicalLabel(text="DATA_OUT", x=10.0, y=20.0, uuid_str="hl-uuid")
        sexp = hl.to_sexp_node()
        assert sexp.name == "hierarchical_label"

    def test_hier_label_to_sexp(self):
        """HierarchicalLabel generates S-expression string."""
        hl = HierarchicalLabel(text="DATA_OUT", x=10.0, y=20.0)
        sexp_str = hl.to_sexp()
        assert "hierarchical_label" in sexp_str
        assert "DATA_OUT" in sexp_str

    def test_hier_label_from_sexp(self):
        """Parse hierarchical label from S-expression node."""
        hl_sexp = SExp.list(
            "hierarchical_label",
            "DATA_OUT",
            SExp.list("shape", "output"),
            SExp.list("at", 10.0, 20.0, 180),
            SExp.list("uuid", "hl-uuid-123"),
        )

        hl = HierarchicalLabel.from_sexp(hl_sexp)
        assert hl.text == "DATA_OUT"
        assert hl.shape == "output"
        assert hl.x == 10.0
        assert hl.y == 20.0
        assert hl.rotation == 180

    def test_hier_label_from_sexp_default_shape(self):
        """Parse hierarchical label defaults to input shape."""
        hl_sexp = SExp.list(
            "hierarchical_label",
            "DATA_IN",
            SExp.list("at", 10.0, 20.0),
            SExp.list("uuid", "hl-uuid"),
        )

        hl = HierarchicalLabel.from_sexp(hl_sexp)
        assert hl.shape == "input"


class TestPowerSymbol:
    """Tests for PowerSymbol dataclass."""

    def test_power_symbol_creation(self):
        """Create power symbol with defaults."""
        pwr = PowerSymbol(lib_id="power:GND", x=10.0, y=20.0)
        assert pwr.lib_id == "power:GND"
        assert pwr.x == 10.0
        assert pwr.y == 20.0
        assert pwr.rotation == 0
        assert pwr.reference == "#PWR?"

    def test_power_symbol_with_reference(self):
        """Create power symbol with specific reference."""
        pwr = PowerSymbol(lib_id="power:+3.3V", x=10.0, y=20.0, reference="#PWR01")
        assert pwr.reference == "#PWR01"

    def test_power_symbol_to_sexp_node(self):
        """PowerSymbol generates valid S-expression node."""
        pwr = PowerSymbol(lib_id="power:GND", x=10.0, y=20.0, uuid_str="pwr-uuid")
        sexp = pwr.to_sexp_node("test_project", "/sheet-uuid")
        assert sexp.name == "symbol"
        # Check lib_id
        lib_id_node = sexp.get("lib_id")
        assert lib_id_node is not None

    def test_power_symbol_to_sexp(self):
        """PowerSymbol generates S-expression string."""
        pwr = PowerSymbol(lib_id="power:GND", x=10.0, y=20.0)
        sexp_str = pwr.to_sexp("test_project", "/sheet-uuid")
        assert "symbol" in sexp_str
        assert "power:GND" in sexp_str

    def test_power_symbol_from_sexp(self):
        """Parse power symbol from S-expression node."""
        pwr_sexp = SExp.list(
            "symbol",
            SExp.list("lib_id", "power:+3.3V"),
            SExp.list("at", 10.0, 20.0, 90),
            SExp.list("uuid", "pwr-uuid-123"),
            SExp.list("property", "Reference", "#PWR05"),
        )

        pwr = PowerSymbol.from_sexp(pwr_sexp)
        assert pwr.lib_id == "power:+3.3V"
        assert pwr.x == 10.0
        assert pwr.y == 20.0
        assert pwr.rotation == 90
        assert pwr.reference == "#PWR05"

    def test_is_power_symbol_by_lib_id(self):
        """Identify power symbol by lib_id."""
        pwr_sexp = SExp.list(
            "symbol",
            SExp.list("lib_id", "power:VCC"),
            SExp.list("at", 10.0, 20.0),
        )
        assert PowerSymbol.is_power_symbol(pwr_sexp) is True

    def test_is_power_symbol_by_reference(self):
        """Identify power symbol by reference."""
        pwr_sexp = SExp.list(
            "symbol",
            SExp.list("lib_id", "some:lib"),
            SExp.list("at", 10.0, 20.0),
            SExp.list("property", "Reference", "#PWR01"),
        )
        assert PowerSymbol.is_power_symbol(pwr_sexp) is True

    def test_is_not_power_symbol(self):
        """Regular symbol is not identified as power symbol."""
        sym_sexp = SExp.list(
            "symbol",
            SExp.list("lib_id", "Device:R"),
            SExp.list("at", 10.0, 20.0),
            SExp.list("property", "Reference", "R1"),
        )
        assert PowerSymbol.is_power_symbol(sym_sexp) is False


class TestPin:
    """Tests for Pin dataclass."""

    def test_pin_creation(self):
        """Create pin with all attributes."""
        pin = Pin(name="VCC", number="1", x=0.0, y=0.0, angle=0.0, length=2.54, pin_type="power_in")
        assert pin.name == "VCC"
        assert pin.number == "1"
        assert pin.x == 0.0
        assert pin.y == 0.0
        assert pin.angle == 0.0
        assert pin.length == 2.54
        assert pin.pin_type == "power_in"

    def test_pin_from_sexp(self):
        """Parse pin from S-expression node."""
        pin_sexp = SExp.list(
            "pin",
            "input",
            "line",
            SExp.list("at", 10.0, 20.0, 180),
            SExp.list("length", 5.0),
            SExp.list("name", "DATA"),
            SExp.list("number", "5"),
        )

        pin = Pin.from_sexp(pin_sexp)
        assert pin.name == "DATA"
        assert pin.number == "5"
        assert pin.x == 10.0
        assert pin.y == 20.0
        assert pin.angle == 180
        assert pin.length == 5.0
        assert pin.pin_type == "input"


class TestSymbolDef:
    """Tests for SymbolDef dataclass."""

    def test_symbol_def_creation(self):
        """Create symbol definition."""
        sym_def = SymbolDef(lib_id="Device:R", name="R", raw_sexp="(symbol ...)")
        assert sym_def.lib_id == "Device:R"
        assert sym_def.name == "R"
        assert sym_def.pins == []

    def test_symbol_def_with_pins(self):
        """Create symbol definition with pins."""
        pins = [
            Pin(name="1", number="1", x=-2.54, y=0, angle=0, length=2.54, pin_type="passive"),
            Pin(name="2", number="2", x=2.54, y=0, angle=180, length=2.54, pin_type="passive"),
        ]
        sym_def = SymbolDef(lib_id="Device:R", name="R", raw_sexp="(symbol ...)", pins=pins)
        assert len(sym_def.pins) == 2

    def test_symbol_def_parse_pins_sexp(self):
        """Parse pins from symbol S-expression."""
        sym_sexp = SExp.list(
            "symbol",
            "Device:R",
            SExp.list(
                "symbol",
                "Device:R_0_1",
                SExp.list(
                    "pin",
                    "passive",
                    "line",
                    SExp.list("at", -2.54, 0, 0),
                    SExp.list("length", 2.54),
                    SExp.list("name", "~"),
                    SExp.list("number", "1"),
                ),
                SExp.list(
                    "pin",
                    "passive",
                    "line",
                    SExp.list("at", 2.54, 0, 180),
                    SExp.list("length", 2.54),
                    SExp.list("name", "~"),
                    SExp.list("number", "2"),
                ),
            ),
        )

        pins = SymbolDef._parse_pins_sexp(sym_sexp)
        assert len(pins) == 2
        assert pins[0].number == "1"
        assert pins[1].number == "2"

    def test_symbol_def_get_embedded_sexp(self):
        """Get embedded S-expression for symbol."""
        sym_sexp = SExp.list(
            "symbol",
            "TestSymbol",
            SExp.list("property", "Reference", "U"),
        )
        sym_def = SymbolDef(
            lib_id="Test:TestSymbol", name="TestSymbol", raw_sexp="", _sexp_node=sym_sexp
        )
        embedded = sym_def.get_embedded_sexp()
        assert "symbol" in embedded


class TestSymbolInstance:
    """Tests for SymbolInstance dataclass."""

    @pytest.fixture
    def mock_symbol_def(self):
        """Create a mock SymbolDef with pins."""
        pins = [
            Pin(name="1", number="1", x=-2.54, y=0, angle=0, length=2.54, pin_type="passive"),
            Pin(name="2", number="2", x=2.54, y=0, angle=180, length=2.54, pin_type="passive"),
        ]
        return SymbolDef(lib_id="Device:R", name="R", raw_sexp="(symbol ...)", pins=pins)

    def test_symbol_instance_creation(self, mock_symbol_def):
        """Create symbol instance."""
        inst = SymbolInstance(
            symbol_def=mock_symbol_def,
            x=100.0,
            y=100.0,
            rotation=0,
            reference="R1",
            value="10k",
        )
        assert inst.reference == "R1"
        assert inst.value == "10k"
        assert inst.x == 100.0
        assert inst.y == 100.0
        assert inst.rotation == 0

    def test_symbol_instance_pin_position(self, mock_symbol_def):
        """Get pin position on placed symbol."""
        inst = SymbolInstance(
            symbol_def=mock_symbol_def,
            x=100.0,
            y=100.0,
            rotation=0,
            reference="R1",
            value="10k",
        )

        pos = inst.pin_position("1")
        # Pin 1 is at (-2.54, 0) relative, so absolute is (100-2.54, 100-0)
        assert abs(pos[0] - 97.46) < 0.01
        assert abs(pos[1] - 100.0) < 0.01

    def test_symbol_instance_pin_position_rotated(self, mock_symbol_def):
        """Get pin position on rotated symbol."""
        inst = SymbolInstance(
            symbol_def=mock_symbol_def,
            x=100.0,
            y=100.0,
            rotation=90,  # 90 degrees rotation
            reference="R1",
            value="10k",
        )

        pos = inst.pin_position("1")
        # At 90 degrees, (-2.54, 0) rotates to (0, -2.54)
        # With Y flip in schematic coords: (100+0, 100-(-2.54)) = (100, 102.54)
        assert abs(pos[0] - 100.0) < 0.01
        assert abs(pos[1] - 102.54) < 0.01

    def test_symbol_instance_pin_position_by_number(self, mock_symbol_def):
        """Get pin position by pin number."""
        inst = SymbolInstance(
            symbol_def=mock_symbol_def,
            x=100.0,
            y=100.0,
            rotation=0,
            reference="R1",
            value="10k",
        )

        # Pin 2 should work by number
        pos = inst.pin_position("2")
        assert pos is not None

    def test_symbol_instance_pin_not_found(self, mock_symbol_def):
        """Error when pin not found."""
        from kicad_tools.schematic.exceptions import PinNotFoundError

        inst = SymbolInstance(
            symbol_def=mock_symbol_def,
            x=100.0,
            y=100.0,
            rotation=0,
            reference="R1",
            value="10k",
        )

        with pytest.raises(PinNotFoundError) as exc:
            inst.pin_position("NONEXISTENT")
        assert "NONEXISTENT" in str(exc.value)

    def test_symbol_instance_all_pin_positions(self, mock_symbol_def):
        """Get all pin positions."""
        inst = SymbolInstance(
            symbol_def=mock_symbol_def,
            x=100.0,
            y=100.0,
            rotation=0,
            reference="R1",
            value="10k",
        )

        positions = inst.all_pin_positions()
        assert "1" in positions
        assert "2" in positions

    def test_symbol_instance_to_sexp_node(self, mock_symbol_def):
        """SymbolInstance generates valid S-expression node."""
        inst = SymbolInstance(
            symbol_def=mock_symbol_def,
            x=100.0,
            y=100.0,
            rotation=0,
            reference="R1",
            value="10k",
            uuid_str="inst-uuid",
        )

        sexp = inst.to_sexp_node("test_project", "/sheet-uuid")
        assert sexp.name == "symbol"

    def test_symbol_instance_to_sexp(self, mock_symbol_def):
        """SymbolInstance generates S-expression string."""
        inst = SymbolInstance(
            symbol_def=mock_symbol_def,
            x=100.0,
            y=100.0,
            rotation=0,
            reference="R1",
            value="10k",
        )

        sexp_str = inst.to_sexp("test_project", "/sheet-uuid")
        assert "symbol" in sexp_str
        assert "R1" in sexp_str
        assert "10k" in sexp_str


class TestSchematicSnapMode:
    """Tests for Schematic snap modes."""

    def test_snap_mode_off(self):
        """SnapMode.OFF preserves coordinates."""
        sch = Schematic(title="Test", snap_mode=SnapMode.OFF)
        coord = sch._snap_coord(10.123, "test")
        assert coord == 10.12  # Just rounded, not snapped

    def test_snap_mode_auto(self):
        """SnapMode.AUTO snaps to grid."""
        sch = Schematic(title="Test", snap_mode=SnapMode.AUTO, grid=2.54)
        coord = sch._snap_coord(10.0, "test")
        assert coord == 10.16  # Snapped to nearest 2.54 grid

    def test_snap_point(self):
        """Snap a point to grid."""
        sch = Schematic(title="Test", snap_mode=SnapMode.AUTO, grid=2.54)
        point = sch._snap_point((10.0, 20.0), "test")
        assert point == (10.16, 20.32)


class TestSchematicConstruction:
    """Tests for Schematic construction and element addition."""

    def test_schematic_creation(self):
        """Create basic schematic."""
        sch = Schematic(title="Test Schematic", date="2025-01", revision="A")
        assert sch.title == "Test Schematic"
        assert sch.date == "2025-01"
        assert sch.revision == "A"
        assert sch.symbols == []
        assert sch.wires == []
        assert sch.junctions == []
        assert sch.labels == []

    def test_schematic_sheet_path(self):
        """Get sheet path."""
        sch = Schematic(title="Test", sheet_uuid="test-sheet-uuid")
        assert "/test-sheet-uuid" in sch.sheet_path

    def test_schematic_sheet_path_with_parent(self):
        """Get sheet path with parent."""
        sch = Schematic(title="Test", sheet_uuid="child-uuid", parent_uuid="parent-uuid")
        assert sch.sheet_path == "/parent-uuid/child-uuid"

    def test_add_wire(self):
        """Add wire to schematic."""
        sch = Schematic(title="Test", snap_mode=SnapMode.OFF)
        wire = sch.add_wire((10, 20), (30, 40), snap=False)
        assert len(sch.wires) == 1
        assert wire.x1 == 10
        assert wire.y1 == 20

    def test_add_wire_path(self):
        """Add connected wire segments."""
        sch = Schematic(title="Test", snap_mode=SnapMode.OFF)
        wires = sch.add_wire_path((0, 0), (10, 0), (10, 10), snap=False)
        assert len(wires) == 2
        assert len(sch.wires) == 2

    def test_add_junction(self):
        """Add junction to schematic."""
        sch = Schematic(title="Test", snap_mode=SnapMode.OFF)
        junc = sch.add_junction(10, 20, snap=False)
        assert len(sch.junctions) == 1
        assert junc.x == 10
        assert junc.y == 20

    def test_add_label(self):
        """Add label to schematic."""
        sch = Schematic(title="Test", snap_mode=SnapMode.OFF)
        label = sch.add_label("NET1", 10, 20, snap=False)
        assert len(sch.labels) == 1
        assert label.text == "NET1"

    def test_add_hier_label(self):
        """Add hierarchical label to schematic."""
        sch = Schematic(title="Test", snap_mode=SnapMode.OFF)
        hl = sch.add_hier_label("DATA", 10, 20, shape="output", snap=False)
        assert len(sch.hier_labels) == 1
        assert hl.text == "DATA"
        assert hl.shape == "output"

    def test_add_text(self):
        """Add text note to schematic."""
        sch = Schematic(title="Test", snap_mode=SnapMode.OFF)
        sch.add_text("Note here", 10, 20, snap=False)
        assert len(sch.text_notes) == 1
        assert sch.text_notes[0] == ("Note here", 10, 20)

    def test_add_rail(self):
        """Add power rail wire."""
        sch = Schematic(title="Test", snap_mode=SnapMode.OFF)
        wire = sch.add_rail(y=100, x_start=10, x_end=50, net_label="VCC", snap=False)
        assert len(sch.wires) == 1
        assert len(sch.labels) == 1
        assert sch.labels[0].text == "VCC"


class TestSchematicQueries:
    """Tests for Schematic query methods."""

    def test_find_label(self):
        """Find label by name."""
        sch = Schematic(title="Test", snap_mode=SnapMode.OFF)
        sch.add_label("NET1", 10, 20, snap=False)
        sch.add_label("NET2", 30, 40, snap=False)

        label = sch.find_label("NET1")
        assert label is not None
        assert label.text == "NET1"

    def test_find_label_not_found(self):
        """Find label returns None when not found."""
        sch = Schematic(title="Test")
        label = sch.find_label("NONEXISTENT")
        assert label is None

    def test_find_labels_pattern(self):
        """Find labels by pattern."""
        sch = Schematic(title="Test", snap_mode=SnapMode.OFF)
        sch.add_label("DATA_IN", 10, 20, snap=False)
        sch.add_label("DATA_OUT", 30, 40, snap=False)
        sch.add_label("CLK", 50, 60, snap=False)

        matches = sch.find_labels("DATA_*")
        assert len(matches) == 2

    def test_find_labels_all(self):
        """Find all labels."""
        sch = Schematic(title="Test", snap_mode=SnapMode.OFF)
        sch.add_label("NET1", 10, 20, snap=False)
        sch.add_label("NET2", 30, 40, snap=False)

        matches = sch.find_labels()
        assert len(matches) == 2

    def test_find_hier_label(self):
        """Find hierarchical label by name."""
        sch = Schematic(title="Test", snap_mode=SnapMode.OFF)
        sch.add_hier_label("DATA", 10, 20, snap=False)

        hl = sch.find_hier_label("DATA")
        assert hl is not None
        assert hl.text == "DATA"

    def test_find_hier_labels_pattern(self):
        """Find hierarchical labels by pattern."""
        sch = Schematic(title="Test", snap_mode=SnapMode.OFF)
        sch.add_hier_label("I2C_SDA", 10, 20, snap=False)
        sch.add_hier_label("I2C_SCL", 30, 40, snap=False)
        sch.add_hier_label("SPI_MOSI", 50, 60, snap=False)

        matches = sch.find_hier_labels("I2C_*")
        assert len(matches) == 2

    def test_find_wires_at_endpoint(self):
        """Find wires by endpoint."""
        sch = Schematic(title="Test", snap_mode=SnapMode.OFF)
        sch.add_wire((10, 20), (30, 20), snap=False)
        sch.add_wire((50, 60), (70, 60), snap=False)

        wires = sch.find_wires(endpoint=(10, 20))
        assert len(wires) == 1

    def test_find_wires_near_point(self):
        """Find wires near a point."""
        sch = Schematic(title="Test", snap_mode=SnapMode.OFF)
        sch.add_wire((10, 20), (30, 20), snap=False)
        sch.add_wire((50, 60), (70, 60), snap=False)

        wires = sch.find_wires(near=(11, 21), tolerance=5)
        assert len(wires) == 1

    def test_find_wires_connected_to_label(self):
        """Find wires connected to a label."""
        sch = Schematic(title="Test", snap_mode=SnapMode.OFF)
        sch.add_label("NET1", 10, 20, snap=False)
        sch.add_wire((10, 20), (30, 20), snap=False)

        wires = sch.find_wires(connected_to_label="NET1")
        assert len(wires) == 1


class TestSchematicRemoval:
    """Tests for Schematic removal methods."""

    def test_remove_wire(self):
        """Remove a wire."""
        sch = Schematic(title="Test", snap_mode=SnapMode.OFF)
        wire = sch.add_wire((10, 20), (30, 40), snap=False)
        assert len(sch.wires) == 1

        result = sch.remove_wire(wire)
        assert result is True
        assert len(sch.wires) == 0

    def test_remove_wire_not_found(self):
        """Remove wire returns False when not found."""
        sch = Schematic(title="Test")
        other_wire = Wire(x1=0, y1=0, x2=10, y2=10)

        result = sch.remove_wire(other_wire)
        assert result is False

    def test_remove_wires_at_point(self):
        """Remove wires at a point."""
        sch = Schematic(title="Test", snap_mode=SnapMode.OFF)
        sch.add_wire((10, 20), (30, 20), snap=False)
        sch.add_wire((10, 20), (10, 40), snap=False)
        assert len(sch.wires) == 2

        count = sch.remove_wires_at((10, 20), tolerance=1)
        assert count == 2
        assert len(sch.wires) == 0

    def test_remove_label(self):
        """Remove label by name."""
        sch = Schematic(title="Test", snap_mode=SnapMode.OFF)
        sch.add_label("NET1", 10, 20, snap=False)
        assert len(sch.labels) == 1

        result = sch.remove_label("NET1")
        assert result is True
        assert len(sch.labels) == 0

    def test_remove_label_not_found(self):
        """Remove label returns False when not found."""
        sch = Schematic(title="Test")
        result = sch.remove_label("NONEXISTENT")
        assert result is False

    def test_remove_hier_label(self):
        """Remove hierarchical label by name."""
        sch = Schematic(title="Test", snap_mode=SnapMode.OFF)
        sch.add_hier_label("DATA", 10, 20, snap=False)

        result = sch.remove_hier_label("DATA")
        assert result is True
        assert len(sch.hier_labels) == 0

    def test_remove_junction(self):
        """Remove junction at position."""
        sch = Schematic(title="Test", snap_mode=SnapMode.OFF)
        sch.add_junction(10, 20, snap=False)

        result = sch.remove_junction(10, 20, tolerance=1)
        assert result is True
        assert len(sch.junctions) == 0

    def test_remove_net(self):
        """Remove net (label and connected wires)."""
        sch = Schematic(title="Test", snap_mode=SnapMode.OFF)
        sch.add_label("NET1", 10, 20, snap=False)
        sch.add_wire((10, 20), (30, 20), snap=False)

        result = sch.remove_net("NET1")
        assert result["label_removed"] is True
        assert result["wires_removed"] == 1
        assert len(sch.labels) == 0
        assert len(sch.wires) == 0


class TestSchematicOutput:
    """Tests for Schematic output methods."""

    def test_to_sexp_node(self):
        """Generate S-expression tree."""
        sch = Schematic(title="Test", snap_mode=SnapMode.OFF)
        sch.add_wire((10, 20), (30, 40), snap=False)
        sch.add_junction(10, 20, snap=False)
        sch.add_label("NET1", 10, 20, snap=False)

        sexp = sch.to_sexp_node()
        assert sexp.name == "kicad_sch"

    def test_to_sexp(self):
        """Generate S-expression string."""
        sch = Schematic(title="Test", snap_mode=SnapMode.OFF)
        sch.add_wire((10, 20), (30, 40), snap=False)

        sexp_str = sch.to_sexp()
        assert "kicad_sch" in sexp_str
        assert "wire" in sexp_str

    def test_write_file(self, tmp_path):
        """Write schematic to file."""
        sch = Schematic(title="Test", snap_mode=SnapMode.OFF)
        sch.add_wire((10, 20), (30, 40), snap=False)

        output_path = tmp_path / "test.kicad_sch"
        sch.write(output_path)

        assert output_path.exists()
        content = output_path.read_text()
        assert "kicad_sch" in content

    def test_get_statistics(self):
        """Get schematic statistics."""
        sch = Schematic(title="Test", snap_mode=SnapMode.OFF)
        sch.add_wire((10, 20), (30, 40), snap=False)
        sch.add_junction(10, 20, snap=False)
        sch.add_label("NET1", 10, 20, snap=False)

        stats = sch.get_statistics()
        assert stats["wire_count"] == 1
        assert stats["junction_count"] == 1
        assert stats["label_count"] == 1


class TestSchematicValidation:
    """Tests for Schematic validation."""

    def test_points_equal(self):
        """Check points equality within tolerance."""
        sch = Schematic(title="Test")
        assert sch._points_equal((10.0, 20.0), (10.0, 20.0)) is True
        assert sch._points_equal((10.0, 20.0), (10.001, 20.001)) is True
        assert sch._points_equal((10.0, 20.0), (11.0, 20.0)) is False

    def test_point_near(self):
        """Check point proximity."""
        sch = Schematic(title="Test")
        assert sch._point_near((10.0, 20.0), (11.0, 20.0), tolerance=2) is True
        assert sch._point_near((10.0, 20.0), (20.0, 20.0), tolerance=2) is False

    def test_point_on_segment_vertical(self):
        """Check if point is on vertical segment."""
        sch = Schematic(title="Test")
        assert sch._point_on_segment((10, 15), (10, 10), (10, 20)) is True
        assert sch._point_on_segment((10, 25), (10, 10), (10, 20)) is False

    def test_point_on_segment_horizontal(self):
        """Check if point is on horizontal segment."""
        sch = Schematic(title="Test")
        assert sch._point_on_segment((15, 10), (10, 10), (20, 10)) is True
        assert sch._point_on_segment((25, 10), (10, 10), (20, 10)) is False

    def test_point_on_segment_diagonal(self):
        """Diagonal segments return False (orthogonal wires only)."""
        sch = Schematic(title="Test")
        # Diagonal segments are not orthogonal - should return False
        assert sch._point_on_segment((15, 15), (10, 10), (20, 20)) is False


class TestSchematicLoadAndParse:
    """Tests for Schematic loading and parsing."""

    def test_load_file_not_found(self, tmp_path):
        """Load raises FileNotFoundError for missing file."""
        with pytest.raises(FileNotFoundError):
            Schematic.load(tmp_path / "nonexistent.kicad_sch")

    def test_load_from_file(self, tmp_path):
        """Load schematic from file."""
        # Create a minimal schematic file
        sch_content = """(kicad_sch
  (version 20231120)
  (generator "test")
  (uuid "test-uuid-123")
  (paper "A4")
  (title_block
    (title "Test Schematic")
    (date "2025-01")
    (rev "B")
    (company "Test Corp")
    (comment 1 "Comment 1")
    (comment 2 "Comment 2")
  )
  (lib_symbols)
  (wire
    (pts (xy 10 20) (xy 30 20))
    (stroke (width 0) (type default))
    (uuid "wire-uuid")
  )
  (junction
    (at 10 20)
    (diameter 0)
    (uuid "junc-uuid")
  )
  (label "NET1"
    (at 10 20 0)
    (effects (font (size 1.27 1.27)))
    (uuid "label-uuid")
  )
  (hierarchical_label "DATA"
    (shape input)
    (at 50 60 0)
    (effects (font (size 1.27 1.27)))
    (uuid "hl-uuid")
  )
  (text "Note"
    (at 100 100 0)
    (effects (font (size 1.27 1.27)))
    (uuid "text-uuid")
  )
  (sheet_instances
    (project "test_project"
      (path "/test-uuid-123"
        (page "1")
      )
    )
  )
)"""
        sch_path = tmp_path / "test.kicad_sch"
        sch_path.write_text(sch_content)

        sch = Schematic.load(sch_path)
        assert sch.title == "Test Schematic"
        assert sch.date == "2025-01"
        assert sch.revision == "B"
        assert sch.company == "Test Corp"
        assert sch.comment1 == "Comment 1"
        assert sch.comment2 == "Comment 2"
        assert len(sch.wires) == 1
        assert len(sch.junctions) == 1
        assert len(sch.labels) == 1
        assert len(sch.hier_labels) == 1
        assert len(sch.text_notes) == 1
        assert sch.project_name == "test_project"

    def test_load_with_power_symbols(self, tmp_path):
        """Load schematic with power symbols."""
        sch_content = """(kicad_sch
  (version 20231120)
  (generator "test")
  (uuid "test-uuid")
  (paper "A4")
  (lib_symbols)
  (symbol
    (lib_id "power:GND")
    (at 100 100 0)
    (uuid "pwr-uuid")
    (property "Reference" "#PWR01" (at 100 100 0) (effects (hide yes)))
    (property "Value" "GND" (at 100 105 0) (effects (font (size 1.27 1.27))))
    (pin "1" (uuid "pin-uuid"))
    (instances (project "test" (path "/" (reference "#PWR01") (unit 1))))
  )
)"""
        sch_path = tmp_path / "power.kicad_sch"
        sch_path.write_text(sch_content)

        sch = Schematic.load(sch_path)
        assert len(sch.power_symbols) == 1
        assert sch.power_symbols[0].lib_id == "power:GND"
        assert sch.power_symbols[0].reference == "#PWR01"
        # Check power counter is updated
        assert sch._pwr_counter == 2


class TestSchematicSnapModeAdvanced:
    """Advanced tests for snap modes."""

    def test_snap_mode_warn(self):
        """SnapMode.WARN warns on off-grid but doesn't snap."""
        import warnings

        sch = Schematic(title="Test", snap_mode=SnapMode.WARN, grid=2.54)

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            coord = sch._snap_coord(10.0, "test")
            # 10.0 is not on 2.54 grid, should warn
            assert len(w) == 1
            assert "Off-grid" in str(w[0].message)
            # But value is not snapped, just rounded
            assert coord == 10.0

    def test_snap_mode_strict(self):
        """SnapMode.STRICT snaps and warns."""
        import warnings

        sch = Schematic(title="Test", snap_mode=SnapMode.STRICT, grid=2.54)

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            coord = sch._snap_coord(10.0, "test")
            # Should warn and snap
            assert len(w) == 1
            assert "Auto-snapping" in str(w[0].message)
            # Value is snapped to nearest grid
            assert coord == 10.16


class TestSchematicWiringHelpers:
    """Tests for Schematic wiring helper methods."""

    @pytest.fixture
    def sch_with_symbols(self):
        """Create schematic with mock symbols for wiring tests."""
        sch = Schematic(title="Test", snap_mode=SnapMode.OFF)

        # Create a mock symbol def with known pin positions
        pins = [
            Pin(
                name="VIN", number="1", x=-5.08, y=2.54, angle=180, length=2.54, pin_type="power_in"
            ),
            Pin(
                name="VOUT", number="2", x=5.08, y=2.54, angle=0, length=2.54, pin_type="power_out"
            ),
            Pin(name="GND", number="3", x=0, y=-5.08, angle=270, length=2.54, pin_type="power_in"),
            Pin(name="EN", number="4", x=-5.08, y=0, angle=180, length=2.54, pin_type="input"),
        ]
        sym_def = SymbolDef(
            lib_id="Regulator_Linear:AP2204K-3.3", name="AP2204K-3.3", raw_sexp="", pins=pins
        )
        sch._symbol_defs["Regulator_Linear:AP2204K-3.3"] = sym_def

        # Add symbol at position
        inst = SymbolInstance(
            symbol_def=sym_def,
            x=100.0,
            y=100.0,
            rotation=0,
            reference="U1",
            value="AP2204K-3.3",
        )
        sch.symbols.append(inst)

        return sch, inst

    def test_route_orthogonal_aligned(self):
        """Route between aligned points (single wire)."""
        sch = Schematic(title="Test", snap_mode=SnapMode.OFF)
        wires = sch._route_orthogonal((10, 20), (50, 20), "auto")
        assert len(wires) == 1
        assert wires[0].x1 == 10
        assert wires[0].x2 == 50

    def test_route_orthogonal_auto(self):
        """Route with auto direction selection."""
        sch = Schematic(title="Test", snap_mode=SnapMode.OFF)
        # Longer vertical distance -> horizontal first
        wires = sch._route_orthogonal((10, 20), (20, 100), "auto")
        assert len(wires) == 2

    def test_route_orthogonal_horizontal_first(self):
        """Route horizontal then vertical."""
        sch = Schematic(title="Test", snap_mode=SnapMode.OFF)
        wires = sch._route_orthogonal((10, 20), (50, 60), "horizontal_first")
        assert len(wires) == 2
        # First wire goes horizontal to (50, 20)
        assert wires[0].x2 == 50
        assert wires[0].y2 == 20

    def test_route_orthogonal_vertical_first(self):
        """Route vertical then horizontal."""
        sch = Schematic(title="Test", snap_mode=SnapMode.OFF)
        wires = sch._route_orthogonal((10, 20), (50, 60), "vertical_first")
        assert len(wires) == 2
        # First wire goes vertical to (10, 60)
        assert wires[0].x2 == 10
        assert wires[0].y2 == 60

    def test_wire_pin_to_point(self, sch_with_symbols):
        """Wire from symbol pin to a point."""
        sch, sym = sch_with_symbols
        wires = sch.wire_pin_to_point(sym, "VOUT", (150, 100))
        # Should create wires from VOUT pin to target
        assert len(wires) >= 1

    def test_wire_pins(self, sch_with_symbols):
        """Wire two symbol pins together."""
        sch, sym1 = sch_with_symbols

        # Add second symbol
        pins2 = [
            Pin(name="1", number="1", x=-2.54, y=0, angle=0, length=2.54, pin_type="passive"),
            Pin(name="2", number="2", x=2.54, y=0, angle=180, length=2.54, pin_type="passive"),
        ]
        sym_def2 = SymbolDef(lib_id="Device:R", name="R", raw_sexp="", pins=pins2)
        sym2 = SymbolInstance(
            symbol_def=sym_def2,
            x=150.0,
            y=100.0,
            rotation=0,
            reference="R1",
            value="10k",
        )
        sch.symbols.append(sym2)

        wires = sch.wire_pins(sym1, "VOUT", sym2, "1")
        assert len(wires) >= 1


class TestSymbolInstanceAdvanced:
    """Advanced tests for SymbolInstance."""

    @pytest.fixture
    def multi_pin_symbol_def(self):
        """Symbol def with multiple pins for testing."""
        pins = [
            Pin(name="VCC", number="1", x=0, y=5.08, angle=90, length=2.54, pin_type="power_in"),
            Pin(name="GND", number="2", x=0, y=-5.08, angle=270, length=2.54, pin_type="power_in"),
            Pin(
                name="DATA",
                number="3",
                x=-5.08,
                y=0,
                angle=180,
                length=2.54,
                pin_type="bidirectional",
            ),
            Pin(name="CLK", number="4", x=5.08, y=0, angle=0, length=2.54, pin_type="input"),
        ]
        return SymbolDef(lib_id="Test:MultiPin", name="MultiPin", raw_sexp="", pins=pins)

    def test_bounding_box(self, multi_pin_symbol_def):
        """Calculate symbol bounding box from pins."""
        inst = SymbolInstance(
            symbol_def=multi_pin_symbol_def,
            x=100.0,
            y=100.0,
            rotation=0,
            reference="U1",
            value="Test",
        )

        bbox = inst.bounding_box(padding=0)
        # Should encompass all pin positions
        assert bbox[0] <= 94.92  # min_x (100 - 5.08)
        assert bbox[2] >= 105.08  # max_x (100 + 5.08)
        assert bbox[1] <= 94.92  # min_y
        assert bbox[3] >= 105.08  # max_y

    def test_bounding_box_with_padding(self, multi_pin_symbol_def):
        """Bounding box includes padding."""
        inst = SymbolInstance(
            symbol_def=multi_pin_symbol_def,
            x=100.0,
            y=100.0,
            rotation=0,
            reference="U1",
            value="Test",
        )

        bbox_no_pad = inst.bounding_box(padding=0)
        bbox_with_pad = inst.bounding_box(padding=5.0)

        assert bbox_with_pad[0] < bbox_no_pad[0]
        assert bbox_with_pad[2] > bbox_no_pad[2]

    def test_bounding_box_no_pins(self):
        """Bounding box for symbol with no pins uses default."""
        sym_def = SymbolDef(lib_id="Test:NoPins", name="NoPins", raw_sexp="", pins=[])
        inst = SymbolInstance(
            symbol_def=sym_def,
            x=100.0,
            y=100.0,
            rotation=0,
            reference="U1",
            value="Test",
        )

        bbox = inst.bounding_box(padding=0)
        # Default half_size is 5.08
        assert abs(bbox[0] - (100 - 5.08)) < 0.01
        assert abs(bbox[2] - (100 + 5.08)) < 0.01

    def test_overlaps_true(self, multi_pin_symbol_def):
        """Symbols overlap when bounding boxes intersect."""
        inst1 = SymbolInstance(
            symbol_def=multi_pin_symbol_def,
            x=100.0,
            y=100.0,
            rotation=0,
            reference="U1",
            value="Test",
        )
        inst2 = SymbolInstance(
            symbol_def=multi_pin_symbol_def,
            x=105.0,  # Close enough to overlap
            y=100.0,
            rotation=0,
            reference="U2",
            value="Test",
        )

        assert inst1.overlaps(inst2, padding=0) is True

    def test_overlaps_false(self, multi_pin_symbol_def):
        """Symbols don't overlap when far apart."""
        inst1 = SymbolInstance(
            symbol_def=multi_pin_symbol_def,
            x=100.0,
            y=100.0,
            rotation=0,
            reference="U1",
            value="Test",
        )
        inst2 = SymbolInstance(
            symbol_def=multi_pin_symbol_def,
            x=200.0,  # Far away
            y=100.0,
            rotation=0,
            reference="U2",
            value="Test",
        )

        assert inst1.overlaps(inst2, padding=0) is False

    def test_pin_position_case_insensitive(self, multi_pin_symbol_def):
        """Pin lookup is case insensitive."""
        inst = SymbolInstance(
            symbol_def=multi_pin_symbol_def,
            x=100.0,
            y=100.0,
            rotation=0,
            reference="U1",
            value="Test",
        )

        # Should find VCC even with lowercase
        pos = inst.pin_position("vcc")
        assert pos is not None


class TestSchematicAutoLayout:
    """Tests for Schematic auto-layout methods."""

    @pytest.fixture
    def sch_with_placed_symbols(self):
        """Create schematic with some placed symbols."""
        sch = Schematic(title="Test", snap_mode=SnapMode.AUTO, grid=2.54)

        pins = [
            Pin(name="1", number="1", x=-5.08, y=0, angle=180, length=2.54, pin_type="passive"),
            Pin(name="2", number="2", x=5.08, y=0, angle=0, length=2.54, pin_type="passive"),
        ]
        sym_def = SymbolDef(lib_id="Device:R", name="R", raw_sexp="", pins=pins)
        sch._symbol_defs["Device:R"] = sym_def

        # Add some symbols
        for i in range(3):
            inst = SymbolInstance(
                symbol_def=sym_def,
                x=100.0 + i * 20,
                y=100.0,
                rotation=0,
                reference=f"R{i + 1}",
                value="10k",
            )
            sch.symbols.append(inst)

        return sch

    def test_find_overlapping_symbols_none(self, sch_with_placed_symbols):
        """No overlaps when symbols are spaced apart."""
        sch = sch_with_placed_symbols
        overlaps = sch.find_overlapping_symbols(padding=2.54)
        # Symbols are 20mm apart, should not overlap with 2.54mm padding
        assert len(overlaps) == 0

    def test_find_overlapping_symbols_found(self):
        """Find overlapping symbols."""
        sch = Schematic(title="Test", snap_mode=SnapMode.OFF)

        pins = [
            Pin(name="1", number="1", x=-5.08, y=0, angle=180, length=2.54, pin_type="passive"),
        ]
        sym_def = SymbolDef(lib_id="Device:R", name="R", raw_sexp="", pins=pins)
        sch._symbol_defs["Device:R"] = sym_def

        # Add overlapping symbols
        inst1 = SymbolInstance(
            symbol_def=sym_def, x=100.0, y=100.0, rotation=0, reference="R1", value="10k"
        )
        inst2 = SymbolInstance(
            symbol_def=sym_def, x=102.0, y=100.0, rotation=0, reference="R2", value="10k"
        )
        sch.symbols.extend([inst1, inst2])

        overlaps = sch.find_overlapping_symbols(padding=2.54)
        assert len(overlaps) == 1
        assert inst1 in overlaps[0]
        assert inst2 in overlaps[0]

    def test_suggest_position_clear(self, sch_with_placed_symbols):
        """Suggest position returns preferred when clear."""
        sch = sch_with_placed_symbols
        # Far from existing symbols
        pos = sch.suggest_position("Device:R", near=(200, 200), avoid_overlaps=True)
        # Should snap to grid near (200, 200)
        assert abs(pos[0] - 200.66) < 0.1 or abs(pos[0] - 198.12) < 0.1

    def test_suggest_position_avoid_overlap(self):
        """Suggest position finds clear space when preferred overlaps."""
        sch = Schematic(title="Test", snap_mode=SnapMode.OFF, grid=2.54)

        pins = [Pin(name="1", number="1", x=-5.08, y=0, angle=180, length=2.54, pin_type="passive")]
        sym_def = SymbolDef(lib_id="Device:R", name="R", raw_sexp="", pins=pins)
        sch._symbol_defs["Device:R"] = sym_def

        # Place symbol at (100, 100)
        inst = SymbolInstance(
            symbol_def=sym_def, x=100.0, y=100.0, rotation=0, reference="R1", value="10k"
        )
        sch.symbols.append(inst)

        # Request position near existing symbol
        pos = sch.suggest_position("Device:R", near=(100, 100), avoid_overlaps=True)
        # Should move away from (100, 100)
        assert pos != (100, 100)

    def test_position_overlaps(self, sch_with_placed_symbols):
        """Check if a position overlaps existing symbols."""
        sch = sch_with_placed_symbols
        sym_def = sch._symbol_defs["Device:R"]

        # Create temp symbol at overlapping position
        temp = SymbolInstance(
            symbol_def=sym_def, x=100.0, y=100.0, rotation=0, reference="_TEMP_", value=""
        )

        assert sch._position_overlaps(temp, padding=2.54) is True

        # Move to non-overlapping position
        temp.x = 200.0
        temp.y = 200.0
        assert sch._position_overlaps(temp, padding=2.54) is False


class TestSchematicValidationAdvanced:
    """Advanced tests for Schematic validation."""

    def test_validate_duplicate_references(self):
        """Validate detects duplicate references."""
        sch = Schematic(title="Test", snap_mode=SnapMode.OFF)

        pins = [Pin(name="1", number="1", x=0, y=0, angle=0, length=2.54, pin_type="passive")]
        sym_def = SymbolDef(lib_id="Device:R", name="R", raw_sexp="", pins=pins)

        # Add two symbols with same reference
        inst1 = SymbolInstance(
            symbol_def=sym_def, x=100.0, y=100.0, rotation=0, reference="R1", value="10k"
        )
        inst2 = SymbolInstance(
            symbol_def=sym_def, x=150.0, y=100.0, rotation=0, reference="R1", value="20k"
        )
        sch.symbols.extend([inst1, inst2])

        issues = sch.validate()
        dup_issues = [i for i in issues if i["type"] == "duplicate_reference"]
        assert len(dup_issues) == 1
        assert "R1" in dup_issues[0]["message"]

    def test_validate_off_grid_symbols_fix(self):
        """Validate can fix off-grid symbols."""
        sch = Schematic(title="Test", snap_mode=SnapMode.OFF, grid=2.54)

        pins = [Pin(name="1", number="1", x=0, y=0, angle=0, length=2.54, pin_type="passive")]
        sym_def = SymbolDef(lib_id="Device:R", name="R", raw_sexp="", pins=pins)

        # Add off-grid symbol
        inst = SymbolInstance(
            symbol_def=sym_def, x=100.1, y=100.2, rotation=0, reference="R1", value="10k"
        )
        sch.symbols.append(inst)

        issues = sch.validate(fix_auto=True)
        off_grid = [i for i in issues if i["type"] == "off_grid_symbol"]
        assert len(off_grid) == 1
        assert off_grid[0]["fix_applied"] is True
        # Symbol should now be on grid
        assert inst.x % 2.54 < 0.01 or abs(inst.x % 2.54 - 2.54) < 0.01

    def test_check_wire_connectivity_floating(self):
        """Detect floating wire endpoints."""
        sch = Schematic(title="Test", snap_mode=SnapMode.OFF)

        # Add wire not connected to anything
        sch.add_wire((100, 100), (150, 100), snap=False)

        issues = sch._check_wire_connectivity()
        floating = [i for i in issues if i["type"] == "floating_wire"]
        assert len(floating) >= 1

    def test_check_wire_connectivity_t_junction(self):
        """Detect T-junction without junction dot."""
        sch = Schematic(title="Test", snap_mode=SnapMode.OFF)

        # Create T-junction: horizontal wire with another wire ending on it
        sch.add_wire((0, 100), (100, 100), snap=False)  # Horizontal
        sch.add_wire((50, 100), (50, 50), snap=False)  # Vertical from middle

        issues = sch._check_wire_connectivity()
        # One end of vertical wire connects to horizontal at (50, 100) - should detect missing junction
        # Other end at (50, 50) is floating
        t_junctions = [i for i in issues if i["type"] == "missing_junction"]
        # Note: May or may not detect depending on implementation

    def test_check_power_pins_connected(self):
        """Power pins connected to wires don't generate warnings."""
        sch = Schematic(title="Test", snap_mode=SnapMode.OFF)

        pins = [Pin(name="VCC", number="1", x=0, y=0, angle=0, length=0, pin_type="power_in")]
        sym_def = SymbolDef(lib_id="Device:IC", name="IC", raw_sexp="", pins=pins)

        inst = SymbolInstance(
            symbol_def=sym_def, x=100.0, y=100.0, rotation=0, reference="U1", value="IC"
        )
        sch.symbols.append(inst)

        # Connect wire to the power pin position
        sch.add_wire((100, 100), (100, 50), snap=False)

        issues = sch._check_power_pins()
        unconnected = [i for i in issues if i["type"] == "unconnected_power_pin"]
        assert len(unconnected) == 0


class TestSymbolDefParsing:
    """Tests for SymbolDef parsing methods."""

    def test_add_prefix_to_node_atom(self):
        """Atom nodes are returned as-is."""
        sym_def = SymbolDef(lib_id="Test:Sym", name="Sym", raw_sexp="")
        atom = SExp.atom("test")
        result = sym_def._add_prefix_to_node(atom, "Lib")
        assert result == atom

    def test_add_prefix_to_node_symbol(self):
        """Symbol names get library prefix."""
        sym_def = SymbolDef(lib_id="Test:Sym", name="Sym", raw_sexp="")
        sym_node = SExp.list("symbol", "TestSymbol", SExp.list("property", "test"))

        result = sym_def._add_prefix_to_node(sym_node, "MyLib")
        # First atom child should now have prefix
        first_child = result.children[0]
        assert str(first_child.value) == "MyLib:TestSymbol"

    def test_add_prefix_to_node_unit_symbol(self):
        """Unit symbols (with _N_N suffix) get prefixed correctly."""
        sym_def = SymbolDef(lib_id="Test:Sym", name="Sym", raw_sexp="")
        sym_node = SExp.list("symbol", "TestSymbol_0_1")

        result = sym_def._add_prefix_to_node(sym_node, "MyLib")
        first_child = result.children[0]
        # Should be prefixed as MyLib:TestSymbol_0_1
        assert str(first_child.value) == "MyLib:TestSymbol_0_1"

    def test_to_sexp_nodes_with_sexp_node(self):
        """Get SExp nodes when _sexp_node is available."""
        sym_node = SExp.list("symbol", "TestSymbol", SExp.list("property", "test"))
        sym_def = SymbolDef(
            lib_id="Test:TestSymbol", name="TestSymbol", raw_sexp="", _sexp_node=sym_node
        )

        nodes = sym_def.to_sexp_nodes()
        assert len(nodes) == 1
        assert nodes[0].name == "symbol"


class TestSymbolInstanceFromSexp:
    """Tests for SymbolInstance.from_sexp parsing."""

    def test_from_sexp_basic(self):
        """Parse basic symbol instance from sexp."""
        sym_sexp = SExp.list(
            "symbol",
            SExp.list("lib_id", "Device:R"),
            SExp.list("at", 100.0, 100.0, 90),
            SExp.list("unit", 1),
            SExp.list("uuid", "test-uuid"),
            SExp.list("property", "Reference", "R1"),
            SExp.list("property", "Value", "10k"),
            SExp.list("property", "Footprint", "Resistor_SMD:R_0402"),
        )

        # Create a mock symbol def
        pins = [Pin(name="1", number="1", x=0, y=0, angle=0, length=2.54, pin_type="passive")]
        mock_def = SymbolDef(lib_id="Device:R", name="R", raw_sexp="", pins=pins)

        inst = SymbolInstance.from_sexp(sym_sexp, symbol_defs={"Device:R": mock_def})

        assert inst.reference == "R1"
        assert inst.value == "10k"
        assert inst.footprint == "Resistor_SMD:R_0402"
        assert inst.x == 100.0
        assert inst.y == 100.0
        assert inst.rotation == 90
        assert inst.unit == 1

    def test_from_sexp_with_embedded_lib_symbols(self):
        """Parse using embedded lib_symbols."""
        sym_sexp = SExp.list(
            "symbol",
            SExp.list("lib_id", "Custom:Part"),
            SExp.list("at", 50.0, 50.0),
            SExp.list("uuid", "inst-uuid"),
            SExp.list("property", "Reference", "U1"),
            SExp.list("property", "Value", "Part"),
        )

        # Embedded lib symbol definition
        lib_sym = SExp.list(
            "symbol",
            "Custom:Part",
            SExp.list(
                "pin",
                "input",
                "line",
                SExp.list("at", 0, 0, 0),
                SExp.list("length", 2.54),
                SExp.list("name", "IN"),
                SExp.list("number", "1"),
            ),
        )

        inst = SymbolInstance.from_sexp(sym_sexp, lib_symbols={"Custom:Part": lib_sym})

        assert inst.reference == "U1"
        assert len(inst.symbol_def.pins) == 1
        assert inst.symbol_def.pins[0].name == "IN"

    def test_from_sexp_placeholder_when_lib_missing(self):
        """Create placeholder SymbolDef when library not found."""
        sym_sexp = SExp.list(
            "symbol",
            SExp.list("lib_id", "Unknown:Missing"),
            SExp.list("at", 50.0, 50.0),
            SExp.list("uuid", "inst-uuid"),
            SExp.list("property", "Reference", "X1"),
            SExp.list("property", "Value", "?"),
        )

        # Disable registry to force fallback behavior
        import kicad_tools.schematic.models.symbol as symbol_module

        original_registry_available = symbol_module._REGISTRY_AVAILABLE
        symbol_module._REGISTRY_AVAILABLE = False

        try:
            # No symbol_defs or lib_symbols provided
            inst = SymbolInstance.from_sexp(sym_sexp)

            # Should create placeholder with no pins
            assert inst.reference == "X1"
            assert inst.symbol_def.lib_id == "Unknown:Missing"
            assert len(inst.symbol_def.pins) == 0
        finally:
            symbol_module._REGISTRY_AVAILABLE = original_registry_available


class TestSchematicSymbolSearch:
    """Tests for symbol search methods."""

    @pytest.fixture
    def sch_with_mixed_symbols(self):
        """Create schematic with various symbols."""
        sch = Schematic(title="Test", snap_mode=SnapMode.OFF)

        pins = [Pin(name="1", number="1", x=0, y=0, angle=0, length=2.54, pin_type="passive")]

        # Add resistors
        r_def = SymbolDef(lib_id="Device:R", name="R", raw_sexp="", pins=pins)
        for i in range(3):
            inst = SymbolInstance(
                symbol_def=r_def,
                x=100.0 + i * 20,
                y=100.0,
                rotation=0,
                reference=f"R{i + 1}",
                value="10k",
            )
            sch.symbols.append(inst)

        # Add capacitors
        c_def = SymbolDef(lib_id="Device:C", name="C", raw_sexp="", pins=pins)
        for i in range(2):
            inst = SymbolInstance(
                symbol_def=c_def,
                x=100.0 + i * 20,
                y=150.0,
                rotation=0,
                reference=f"C{i + 1}",
                value="100nF",
            )
            sch.symbols.append(inst)

        return sch

    def test_find_symbol(self, sch_with_mixed_symbols):
        """Find symbol by reference."""
        sch = sch_with_mixed_symbols
        sym = sch.find_symbol("R2")
        assert sym is not None
        assert sym.reference == "R2"

    def test_find_symbol_not_found(self, sch_with_mixed_symbols):
        """Find symbol returns None when not found."""
        sch = sch_with_mixed_symbols
        sym = sch.find_symbol("X99")
        assert sym is None

    def test_find_symbols_pattern(self, sch_with_mixed_symbols):
        """Find symbols by pattern."""
        sch = sch_with_mixed_symbols
        resistors = sch.find_symbols("R*")
        assert len(resistors) == 3

    def test_find_symbols_all(self, sch_with_mixed_symbols):
        """Find all symbols."""
        sch = sch_with_mixed_symbols
        all_syms = sch.find_symbols()
        assert len(all_syms) == 5

    def test_find_symbols_by_value(self, sch_with_mixed_symbols):
        """Find symbols by value."""
        sch = sch_with_mixed_symbols
        found = sch.find_symbols_by_value("10k")
        assert len(found) == 3

    def test_find_symbols_by_lib(self, sch_with_mixed_symbols):
        """Find symbols by library pattern."""
        sch = sch_with_mixed_symbols
        found = sch.find_symbols_by_lib("Device:C")
        assert len(found) == 2

    def test_remove_symbol(self, sch_with_mixed_symbols):
        """Remove symbol by reference."""
        sch = sch_with_mixed_symbols
        assert len(sch.symbols) == 5

        result = sch.remove_symbol("R2")
        assert result is True
        assert len(sch.symbols) == 4
        assert sch.find_symbol("R2") is None

    def test_remove_symbol_not_found(self, sch_with_mixed_symbols):
        """Remove symbol returns False when not found."""
        sch = sch_with_mixed_symbols
        result = sch.remove_symbol("X99")
        assert result is False



"""Tests for the keepout zone management module."""

import tempfile

from kicad_tools.optim import (
    KeepoutType,
    KeepoutViolation,
    KeepoutZone,
    PlacementOptimizer,
    Polygon,
    add_keepout_zones,
    create_keepout_polygon,
    load_keepout_zones_from_yaml,
)


class TestKeepoutType:
    """Tests for KeepoutType enum."""

    def test_enum_values(self):
        assert KeepoutType.MECHANICAL.value == "mechanical"
        assert KeepoutType.THERMAL.value == "thermal"
        assert KeepoutType.RF.value == "rf"
        assert KeepoutType.ASSEMBLY.value == "assembly"
        assert KeepoutType.CLEARANCE.value == "clearance"

    def test_from_string(self):
        assert KeepoutType("mechanical") == KeepoutType.MECHANICAL
        assert KeepoutType("thermal") == KeepoutType.THERMAL
        assert KeepoutType("rf") == KeepoutType.RF


class TestKeepoutZone:
    """Tests for KeepoutZone dataclass."""

    def test_basic_zone_creation(self):
        zone = KeepoutZone(
            name="test_zone",
            zone_type=KeepoutType.MECHANICAL,
            polygon=[(0, 0), (10, 0), (10, 10), (0, 10)],
        )
        assert zone.name == "test_zone"
        assert zone.zone_type == KeepoutType.MECHANICAL
        assert len(zone.polygon) == 4

    def test_zone_with_clearance(self):
        zone = KeepoutZone(
            name="clearance_zone",
            zone_type=KeepoutType.CLEARANCE,
            polygon=[(0, 0), (10, 0), (10, 10), (0, 10)],
            clearance_mm=2.0,
        )
        assert zone.clearance_mm == 2.0

    def test_zone_layer_restriction(self):
        zone = KeepoutZone(
            name="top_zone",
            zone_type=KeepoutType.MECHANICAL,
            polygon=[(0, 0), (10, 0), (10, 10), (0, 10)],
            layer="F.Cu",
        )
        assert zone.layer == "F.Cu"

    def test_zone_via_trace_permissions(self):
        zone = KeepoutZone(
            name="rf_zone",
            zone_type=KeepoutType.RF,
            polygon=[(0, 0), (10, 0), (10, 10), (0, 10)],
            allow_vias=True,
            allow_traces=False,
        )
        assert zone.allow_vias is True
        assert zone.allow_traces is False

    def test_get_polygon(self):
        zone = KeepoutZone(
            name="test",
            zone_type=KeepoutType.MECHANICAL,
            polygon=[(0, 0), (10, 0), (10, 10), (0, 10)],
        )
        poly = zone.get_polygon()
        assert isinstance(poly, Polygon)
        assert len(poly.vertices) == 4

    def test_get_expanded_polygon(self):
        zone = KeepoutZone(
            name="test",
            zone_type=KeepoutType.MECHANICAL,
            polygon=[(0, 0), (10, 0), (10, 10), (0, 10)],
            clearance_mm=1.0,
        )
        poly = zone.get_expanded_polygon()
        # Expanded polygon should be larger
        min_x = min(v.x for v in poly.vertices)
        max_x = max(v.x for v in poly.vertices)
        # The expansion should push vertices outward
        assert min_x < 0  # Expanded beyond original boundary
        assert max_x > 10

    def test_contains_point_inside(self):
        zone = KeepoutZone(
            name="test",
            zone_type=KeepoutType.MECHANICAL,
            polygon=[(0, 0), (10, 0), (10, 10), (0, 10)],
        )
        assert zone.contains_point(5.0, 5.0) is True

    def test_contains_point_outside(self):
        zone = KeepoutZone(
            name="test",
            zone_type=KeepoutType.MECHANICAL,
            polygon=[(0, 0), (10, 0), (10, 10), (0, 10)],
        )
        assert zone.contains_point(15.0, 15.0) is False

    def test_contains_point_with_clearance(self):
        zone = KeepoutZone(
            name="test",
            zone_type=KeepoutType.MECHANICAL,
            polygon=[(0, 0), (10, 0), (10, 10), (0, 10)],
            clearance_mm=2.0,
        )
        # Point at (11, 5) is outside the base polygon but inside clearance
        # Due to expansion algorithm, points near edges should be included
        # The center should definitely be inside
        assert zone.contains_point(5.0, 5.0) is True

    def test_to_keepout(self):
        zone = KeepoutZone(
            name="test_zone",
            zone_type=KeepoutType.MECHANICAL,
            polygon=[(0, 0), (10, 0), (10, 10), (0, 10)],
            charge_multiplier=15.0,
        )
        keepout = zone.to_keepout()
        assert keepout.name == "test_zone"
        assert keepout.charge_multiplier == 15.0
        assert len(keepout.outline.vertices) == 4

    def test_to_dict(self):
        zone = KeepoutZone(
            name="test",
            zone_type=KeepoutType.RF,
            polygon=[(0, 0), (10, 0), (10, 10), (0, 10)],
            clearance_mm=3.0,
            layer="B.Cu",
        )
        d = zone.to_dict()
        assert d["name"] == "test"
        assert d["type"] == "rf"
        assert d["clearance_mm"] == 3.0
        assert d["layer"] == "B.Cu"

    def test_from_dict(self):
        data = {
            "name": "antenna",
            "type": "rf",
            "polygon": [[0, 0], [20, 0], [20, 15], [0, 15]],
            "clearance_mm": 5.0,
            "allow_traces": True,
        }
        zone = KeepoutZone.from_dict(data)
        assert zone.name == "antenna"
        assert zone.zone_type == KeepoutType.RF
        assert zone.clearance_mm == 5.0
        assert zone.allow_traces is True


class TestKeepoutViolation:
    """Tests for KeepoutViolation dataclass."""

    def test_violation_creation(self):
        violation = KeepoutViolation(
            component_ref="U1",
            zone_name="rf_zone",
            zone_type=KeepoutType.RF,
            position=(50.0, 25.0),
            overlap_mm=2.5,
            message="U1 overlaps RF exclusion zone",
        )
        assert violation.component_ref == "U1"
        assert violation.zone_name == "rf_zone"
        assert violation.overlap_mm == 2.5

    def test_violation_to_dict(self):
        violation = KeepoutViolation(
            component_ref="C3",
            zone_name="thermal_zone",
            zone_type=KeepoutType.THERMAL,
            position=(10.0, 20.0),
            overlap_mm=1.0,
        )
        d = violation.to_dict()
        assert d["component"] == "C3"
        assert d["zone"] == "thermal_zone"
        assert d["type"] == "thermal"


class TestCreateKeepoutPolygon:
    """Tests for create_keepout_polygon helper."""

    def test_basic_polygon(self):
        vertices = [(0, 0), (10, 0), (10, 10), (0, 10)]
        zone = create_keepout_polygon(vertices, KeepoutType.MECHANICAL)
        assert len(zone.polygon) == 4
        assert zone.zone_type == KeepoutType.MECHANICAL

    def test_polygon_with_name(self):
        vertices = [(0, 0), (5, 0), (5, 5)]
        zone = create_keepout_polygon(vertices, KeepoutType.RF, name="antenna_clearance")
        assert zone.name == "antenna_clearance"

    def test_polygon_with_clearance(self):
        vertices = [(0, 0), (10, 0), (10, 10), (0, 10)]
        zone = create_keepout_polygon(vertices, KeepoutType.THERMAL, clearance_mm=2.5)
        assert zone.clearance_mm == 2.5

    def test_polygon_with_layer(self):
        vertices = [(0, 0), (10, 0), (10, 10), (0, 10)]
        zone = create_keepout_polygon(vertices, KeepoutType.MECHANICAL, layer="B.Cu")
        assert zone.layer == "B.Cu"


class TestAddKeepoutZones:
    """Tests for add_keepout_zones function."""

    def test_add_single_zone(self):
        board = Polygon.rectangle(50, 40, 100, 80)
        optimizer = PlacementOptimizer(board)

        zone = KeepoutZone(
            name="test",
            zone_type=KeepoutType.MECHANICAL,
            polygon=[(10, 10), (20, 10), (20, 20), (10, 20)],
        )

        count = add_keepout_zones(optimizer, [zone])
        assert count == 1
        assert len(optimizer.keepouts) == 1

    def test_add_multiple_zones(self):
        board = Polygon.rectangle(50, 40, 100, 80)
        optimizer = PlacementOptimizer(board)

        zones = [
            KeepoutZone(
                name="zone1",
                zone_type=KeepoutType.MECHANICAL,
                polygon=[(0, 0), (10, 0), (10, 10), (0, 10)],
            ),
            KeepoutZone(
                name="zone2",
                zone_type=KeepoutType.RF,
                polygon=[(80, 60), (95, 60), (95, 75), (80, 75)],
            ),
        ]

        count = add_keepout_zones(optimizer, zones)
        assert count == 2
        assert len(optimizer.keepouts) == 2

    def test_zones_affect_forces(self):
        board = Polygon.rectangle(50, 40, 100, 80)
        optimizer = PlacementOptimizer(board)

        # Add a component near where we'll put a keepout
        from kicad_tools.optim import Component

        comp = Component(ref="U1", x=15, y=15, width=5, height=5)
        optimizer.add_component(comp)

        # Add a keepout zone right where the component is
        zone = KeepoutZone(
            name="obstacle",
            zone_type=KeepoutType.MECHANICAL,
            polygon=[(10, 10), (20, 10), (20, 20), (10, 20)],
            charge_multiplier=20.0,
        )
        add_keepout_zones(optimizer, [zone])

        # Compute forces - there should be repulsion
        forces, torques = optimizer.compute_forces_and_torques()
        force = forces["U1"]
        # The component should be pushed away from the keepout
        assert force.magnitude() > 0


class TestLoadKeepoutZonesFromYaml:
    """Tests for load_keepout_zones_from_yaml function."""

    def test_load_single_zone(self):
        yaml_content = """
keepouts:
  - name: usb_clearance
    type: mechanical
    polygon: [[0, 0], [10, 0], [10, 5], [0, 5]]
    clearance_mm: 1.0
"""
        with tempfile.NamedTemporaryFile(mode="w", suffix=".yaml", delete=False) as f:
            f.write(yaml_content)
            f.flush()

            zones = load_keepout_zones_from_yaml(f.name)

        assert len(zones) == 1
        assert zones[0].name == "usb_clearance"
        assert zones[0].zone_type == KeepoutType.MECHANICAL
        assert zones[0].clearance_mm == 1.0

    def test_load_multiple_zones(self):
        yaml_content = """
keepouts:
  - name: usb_zone
    type: mechanical
    polygon: [[0, 0], [10, 0], [10, 5], [0, 5]]
  - name: antenna_zone
    type: rf
    polygon: [[50, 0], [60, 0], [60, 20], [50, 20]]
    clearance_mm: 5.0
  - name: hot_spot
    type: thermal
    polygon: [[30, 30], [40, 30], [40, 40], [30, 40]]
"""
        with tempfile.NamedTemporaryFile(mode="w", suffix=".yaml", delete=False) as f:
            f.write(yaml_content)
            f.flush()

            zones = load_keepout_zones_from_yaml(f.name)

        assert len(zones) == 3
        assert zones[0].zone_type == KeepoutType.MECHANICAL
        assert zones[1].zone_type == KeepoutType.RF
        assert zones[2].zone_type == KeepoutType.THERMAL

    def test_load_with_all_options(self):
        yaml_content = """
keepouts:
  - name: complex_zone
    type: assembly
    polygon: [[0, 0], [20, 0], [20, 20], [0, 20]]
    layer: F.Cu
    clearance_mm: 2.5
    allow_vias: true
    allow_traces: false
    charge_multiplier: 15.0
"""
        with tempfile.NamedTemporaryFile(mode="w", suffix=".yaml", delete=False) as f:
            f.write(yaml_content)
            f.flush()

            zones = load_keepout_zones_from_yaml(f.name)

        assert len(zones) == 1
        zone = zones[0]
        assert zone.zone_type == KeepoutType.ASSEMBLY
        assert zone.layer == "F.Cu"
        assert zone.clearance_mm == 2.5
        assert zone.allow_vias is True
        assert zone.allow_traces is False
        assert zone.charge_multiplier == 15.0

    def test_load_empty_file(self):
        yaml_content = "keepouts: []"
        with tempfile.NamedTemporaryFile(mode="w", suffix=".yaml", delete=False) as f:
            f.write(yaml_content)
            f.flush()

            zones = load_keepout_zones_from_yaml(f.name)

        assert len(zones) == 0


class TestOptimizerIntegration:
    """Integration tests for keepout zones with the optimizer."""

    def test_optimizer_respects_keepout(self):
        """Test that optimizer moves components away from keepout zones."""
        # Create a board
        board = Polygon.rectangle(50, 40, 100, 80)
        optimizer = PlacementOptimizer(board)

        # Add component inside a keepout zone
        from kicad_tools.optim import Component

        comp = Component(ref="U1", x=15, y=15, width=4, height=4)
        optimizer.add_component(comp)

        # Add keepout zone where the component is
        zone = KeepoutZone(
            name="obstacle",
            zone_type=KeepoutType.MECHANICAL,
            polygon=[(10, 10), (25, 10), (25, 25), (10, 25)],
            charge_multiplier=50.0,
        )
        add_keepout_zones(optimizer, [zone])

        # Record initial position
        initial_x, initial_y = comp.x, comp.y

        # Run a few optimization iterations
        optimizer.run(iterations=50, dt=0.05)

        # Component should have moved away from keepout
        moved = (comp.x != initial_x) or (comp.y != initial_y)
        assert moved, "Component should move away from keepout zone"

    def test_multiple_keepouts(self):
        """Test optimizer with multiple keepout zones."""
        board = Polygon.rectangle(50, 40, 100, 80)
        optimizer = PlacementOptimizer(board)

        from kicad_tools.optim import Component

        comp = Component(ref="U1", x=50, y=40, width=5, height=5)
        optimizer.add_component(comp)

        # Add keepouts on all sides
        zones = [
            KeepoutZone(
                name="left",
                zone_type=KeepoutType.MECHANICAL,
                polygon=[(0, 0), (20, 0), (20, 80), (0, 80)],
            ),
            KeepoutZone(
                name="right",
                zone_type=KeepoutType.MECHANICAL,
                polygon=[(80, 0), (100, 0), (100, 80), (80, 80)],
            ),
        ]
        add_keepout_zones(optimizer, zones)

        assert len(optimizer.keepouts) == 2

        # Run optimization
        optimizer.run(iterations=100, dt=0.02)

        # Component should stay in the middle region
        assert 20 < comp.x < 80



#!/usr/bin/env python3
"""
KiCad Symbol Generator

Creates KiCad symbol library files from pin definitions.

Supported input formats:
- JSON pin definition file
- CSV with pin number, name, type columns
- Datasheet-style text (parsed interactively)

Features:
- Automatic pin type detection from names (VCC→power_in, GND→power_in, etc.)
- Intelligent pin arrangement (power top/bottom, I/O left/right)
- Package-aware layouts (DIP, SOIC, QFP, BGA)
- Property generation (Reference, Value, Footprint, Description)

Usage:
    # From JSON definition
    python create_symbol.py --json pins.json --output MyLib.kicad_sym

    # From CSV
    python create_symbol.py --csv pins.csv --name "TPA3116D2" --output Amplifier.kicad_sym

    # Interactive from datasheet text
    python create_symbol.py --interactive --name "PCM5122" --output DAC.kicad_sym

    # From template
    python create_symbol.py --template qfp48 --name "STM32F103" --output MCU.kicad_sym

Example JSON format:
    {
        "name": "TPA3116D2",
        "reference": "U",
        "footprint": "Package_SO:HTSSOP-28-1EP_4.4x9.7mm_P0.65mm_EP3.4x9.5mm",
        "description": "Class-D Audio Amplifier, 2x50W",
        "datasheet": "https://www.ti.com/lit/ds/symlink/tpa3116d2.pdf",
        "pins": [
            {"number": "1", "name": "PVCC", "type": "power_in"},
            {"number": "2", "name": "OUTL+", "type": "output"},
            ...
        ]
    }
"""

import argparse
import contextlib
import csv
import json
import re
from dataclasses import dataclass
from enum import Enum
from pathlib import Path

from kicad_tools.utils import ensure_parent_dir


class PinType(Enum):
    """KiCad pin electrical types."""

    INPUT = "input"
    OUTPUT = "output"
    BIDIRECTIONAL = "bidirectional"
    TRI_STATE = "tri_state"
    PASSIVE = "passive"
    FREE = "free"
    UNSPECIFIED = "unspecified"
    POWER_IN = "power_in"
    POWER_OUT = "power_out"
    OPEN_COLLECTOR = "open_collector"
    OPEN_EMITTER = "open_emitter"
    UNCONNECTED = "unconnected"
    NO_CONNECT = "no_connect"


class PinStyle(Enum):
    """KiCad pin graphic styles."""

    LINE = "line"
    INVERTED = "inverted"
    CLOCK = "clock"
    INVERTED_CLOCK = "inverted_clock"
    INPUT_LOW = "input_low"
    CLOCK_LOW = "clock_low"
    OUTPUT_LOW = "output_low"
    EDGE_CLOCK_HIGH = "edge_clock_high"
    NON_LOGIC = "non_logic"


class PinSide(Enum):
    """Which side of the symbol the pin appears on."""

    LEFT = "left"
    RIGHT = "right"
    TOP = "top"
    BOTTOM = "bottom"


@dataclass
class PinDef:
    """Pin definition for symbol generation."""

    number: str
    name: str
    pin_type: PinType = PinType.PASSIVE
    style: PinStyle = PinStyle.LINE
    side: PinSide | None = None  # Auto-assigned if None
    hidden: bool = False

    @classmethod
    def from_dict(cls, d: dict) -> "PinDef":
        """Create PinDef from dictionary."""
        pin_type = d.get("type", d.get("pin_type", "passive"))
        if isinstance(pin_type, str):
            pin_type = PinType(pin_type.lower())

        style = d.get("style", "line")
        if isinstance(style, str):
            style = PinStyle(style.lower())

        side = d.get("side")
        if side and isinstance(side, str):
            side = PinSide(side.lower())

        return cls(
            number=str(d["number"]),
            name=d.get("name", ""),
            pin_type=pin_type,
            style=style,
            side=side,
            hidden=d.get("hidden", False),
        )


@dataclass
class SymbolDef:
    """Complete symbol definition."""

    name: str
    pins: list[PinDef]
    reference: str = "U"
    value: str = ""
    footprint: str = ""
    datasheet: str = ""
    description: str = ""
    keywords: str = ""

    # Layout options
    pin_length: float = 2.54  # mm
    pin_spacing: float = 2.54  # mm (100 mil)
    min_width: float = 10.16  # mm (400 mil)

    def __post_init__(self):
        if not self.value:
            self.value = self.name


# Pin type detection patterns
PIN_TYPE_PATTERNS = [
    # Power input
    (r"^V(CC|DD|BAT|IN|BUS|IO|REF|A|D|S|L)[\dA-Z]*$", PinType.POWER_IN),
    (r"^(PVCC|AVCC|DVCC|AVDD|DVDD|PVDD)[\dA-Z]*$", PinType.POWER_IN),
    (r"^[+-]?(3V3|5V|12V|1V8|2V5|VCC|VDD)", PinType.POWER_IN),
    (r"^(GND|VSS|AGND|DGND|PGND|GNDA|GNDD)[\dA-Z]*$", PinType.POWER_IN),
    (r"^(EP|EPAD|PAD|THERMAL)$", PinType.POWER_IN),  # Exposed pad
    # No connect
    (r"^N/?C[\dA-Z]*$", PinType.NO_CONNECT),
    (r"^(NC|DNC|RSVD|RESERVED)[\dA-Z]*$", PinType.NO_CONNECT),
    # Clock inputs (with clock style)
    (r"^(CLK|CLOCK|SCK|SCLK|BCLK|MCLK|LRCLK|WS)", PinType.INPUT),
    (r"^X(TAL)?IN[\dA-Z]*$", PinType.INPUT),
    # Reset (active low)
    (r"^(~?RST|~?RESET|~?NRST|RST_?N|RESET_?N)", PinType.INPUT),
    # Standard inputs
    (r"^(SDA|SPI_?MOSI|DIN|DATA_?IN|RX|RXD|D\d*IN)", PinType.INPUT),
    (r"^(EN|ENABLE|~?SHDN|~?SHUTDOWN|~?MUTE|~?STBY|STANDBY)", PinType.INPUT),
    (r"^(GAIN\d*|MODE\d*|SEL\d*|ADR\d*|ADDR\d*)", PinType.INPUT),
    # Standard outputs
    (r"^(SCL|SPI_?MISO|DOUT|DATA_?OUT|TX|TXD|D\d*OUT)", PinType.OUTPUT),
    (r"^(OUT[LRP]?[+-]?|OUTP|OUTN|VOUT|AOUT|DOUT)", PinType.OUTPUT),
    (r"^(FAULT|~?FAULT|ERR|~?ERR|FLAG|INT|~?INT)", PinType.OUTPUT),
    # Bidirectional
    (r"^(GPIO\d*|P[A-Z]?\d+|IO\d*|D\d+)", PinType.BIDIRECTIONAL),
    (r"^(I2C_?SDA|SDA\d*)", PinType.BIDIRECTIONAL),
    # Audio/analog
    (r"^(INL|INR|LINEIN|RINEIN|L_?IN|R_?IN)", PinType.INPUT),
    (r"^(OUTL|OUTR|LINEOUT|ROUTOUT|L_?OUT|R_?OUT|HP)", PinType.OUTPUT),
    (r"^(AIN\d*|ADC\d*)", PinType.INPUT),
    (r"^(AOUT\d*|DAC\d*)", PinType.OUTPUT),
    # Crystal oscillator
    (r"^X(TAL)?OUT[\dA-Z]*$", PinType.OUTPUT),
]

# Pin side assignment patterns (for intelligent layout)
PIN_SIDE_PATTERNS = [
    # Power on top and bottom
    (r"^V(CC|DD|BAT|IN|BUS|IO|REF|A|D|S|L)[\dA-Z]*$", PinSide.TOP),
    (r"^(PVCC|AVCC|DVCC|AVDD|DVDD|PVDD)[\dA-Z]*$", PinSide.TOP),
    (r"^[+-]?(3V3|5V|12V|1V8|2V5|VCC|VDD)", PinSide.TOP),
    (r"^(GND|VSS|AGND|DGND|PGND|GNDA|GNDD)[\dA-Z]*$", PinSide.BOTTOM),
    (r"^(EP|EPAD|PAD|THERMAL)$", PinSide.BOTTOM),
    # Inputs on left
    (r"^(IN|INPUT|SDA|MOSI|DIN|RX|CLK|SCK|EN|RST|~?RST|NRST)", PinSide.LEFT),
    (r"^(GAIN|MODE|SEL|ADR|ADDR|LINEIN|RINEIN|AIN|ADC)", PinSide.LEFT),
    # Outputs on right
    (r"^(OUT|OUTPUT|SCL|MISO|DOUT|TX|FAULT|ERR|FLAG|INT)", PinSide.RIGHT),
    (r"^(LINEOUT|ROUTOUT|HP|AOUT|DAC)", PinSide.RIGHT),
]

# Patterns that indicate clock style pins
CLOCK_PATTERNS = [
    r"^(CLK|CLOCK|SCK|SCLK|BCLK|MCLK|LRCLK|WS|TCK)",
    r"^X(TAL)?IN[\dA-Z]*$",
]

# Patterns that indicate inverted (active low) pins
INVERTED_PATTERNS = [
    r"^~",  # KiCad tilde notation
    r"_N$",  # Suffix _N
    r"N$",  # Suffix N (for RST_N, CS_N)
    r"^/(RST|CS|SS|EN|INT|IRQ)",  # /RST notation
]


def detect_pin_type(name: str) -> PinType:
    """Detect pin type from name using pattern matching."""
    name_upper = name.upper().replace("~", "").replace("/", "")

    for pattern, pin_type in PIN_TYPE_PATTERNS:
        if re.match(pattern, name_upper, re.IGNORECASE):
            return pin_type

    return PinType.PASSIVE


def detect_pin_side(name: str, pin_type: PinType) -> PinSide:
    """Detect which side of symbol a pin should be on."""
    name_upper = name.upper().replace("~", "").replace("/", "")

    # Check explicit patterns first
    for pattern, side in PIN_SIDE_PATTERNS:
        if re.match(pattern, name_upper, re.IGNORECASE):
            return side

    # Fall back to type-based assignment
    if pin_type == PinType.POWER_IN:
        if "GND" in name_upper or "VSS" in name_upper:
            return PinSide.BOTTOM
        return PinSide.TOP
    elif pin_type == PinType.INPUT:
        return PinSide.LEFT
    elif (
        pin_type == PinType.OUTPUT
        or pin_type == PinType.BIDIRECTIONAL
        or pin_type == PinType.NO_CONNECT
    ):
        return PinSide.RIGHT

    return PinSide.LEFT


def detect_pin_style(name: str, pin_type: PinType) -> PinStyle:
    """Detect pin graphic style from name."""
    # Check for clock patterns
    for pattern in CLOCK_PATTERNS:
        if re.match(pattern, name.upper(), re.IGNORECASE):
            return PinStyle.CLOCK

    # Check for inverted (active low) patterns
    for pattern in INVERTED_PATTERNS:
        if re.search(pattern, name, re.IGNORECASE):
            return PinStyle.INVERTED

    return PinStyle.LINE


def parse_json(path: Path) -> SymbolDef:
    """Parse symbol definition from JSON file."""
    with open(path) as f:
        data = json.load(f)

    pins = []
    for p in data.get("pins", []):
        pin = PinDef.from_dict(p)

        # Auto-detect type and side if not specified
        if pin.pin_type == PinType.PASSIVE and "type" not in p:
            pin.pin_type = detect_pin_type(pin.name)
        if pin.side is None:
            pin.side = detect_pin_side(pin.name, pin.pin_type)
        if pin.style == PinStyle.LINE and "style" not in p:
            pin.style = detect_pin_style(pin.name, pin.pin_type)

        pins.append(pin)

    return SymbolDef(
        name=data.get("name", "Unknown"),
        pins=pins,
        reference=data.get("reference", "U"),
        value=data.get("value", data.get("name", "")),
        footprint=data.get("footprint", ""),
        datasheet=data.get("datasheet", ""),
        description=data.get("description", ""),
        keywords=data.get("keywords", ""),
    )


def parse_csv(path: Path, name: str) -> SymbolDef:
    """Parse pin definitions from CSV file.

    Expected columns: number, name, [type], [side], [style]
    """
    pins = []

    with open(path, newline="") as f:
        reader = csv.DictReader(f)
        for row in reader:
            # Normalize column names
            row = {k.lower().strip(): v.strip() for k, v in row.items()}

            pin_type = PinType.PASSIVE
            if "type" in row and row["type"]:
                try:
                    pin_type = PinType(row["type"].lower())
                except ValueError:
                    pin_type = detect_pin_type(row.get("name", ""))
            else:
                pin_type = detect_pin_type(row.get("name", ""))

            side = None
            if "side" in row and row["side"]:
                with contextlib.suppress(ValueError):
                    side = PinSide(row["side"].lower())
            if side is None:
                side = detect_pin_side(row.get("name", ""), pin_type)

            style = PinStyle.LINE
            if "style" in row and row["style"]:
                with contextlib.suppress(ValueError):
                    style = PinStyle(row["style"].lower())
            else:
                style = detect_pin_style(row.get("name", ""), pin_type)

            pins.append(
                PinDef(
                    number=row.get("number", row.get("pin", "")),
                    name=row.get("name", ""),
                    pin_type=pin_type,
                    side=side,
                    style=style,
                )
            )

    return SymbolDef(name=name, pins=pins)


def parse_datasheet_text(text: str, name: str) -> SymbolDef:
    """Parse pin definitions from datasheet-style text.

    Accepts formats like:
        1  PVCC     Power supply
        2  OUT+     Output positive
        3  GND      Ground

    Or tab/comma separated.
    """
    pins = []

    for line in text.strip().split("\n"):
        line = line.strip()
        if not line or line.startswith("#"):
            continue

        # Try different separators
        parts = None
        for sep in ["\t", ",", "  ", " "]:
            parts = [p.strip() for p in line.split(sep) if p.strip()]
            if len(parts) >= 2:
                break

        if not parts or len(parts) < 2:
            continue

        # First part is pin number, second is name
        pin_num = parts[0]
        pin_name = parts[1]

        # Optional third part is description/type hint
        type_hint = parts[2] if len(parts) > 2 else ""

        pin_type = detect_pin_type(pin_name)
        if type_hint:
            # Try to extract type from description
            type_hint_lower = type_hint.lower()
            if (
                "power" in type_hint_lower
                or "supply" in type_hint_lower
                or "ground" in type_hint_lower
            ):
                pin_type = PinType.POWER_IN
            elif "input" in type_hint_lower:
                pin_type = PinType.INPUT
            elif "output" in type_hint_lower:
                pin_type = PinType.OUTPUT
            elif "bidirectional" in type_hint_lower or "i/o" in type_hint_lower:
                pin_type = PinType.BIDIRECTIONAL
            elif "no connect" in type_hint_lower or "n/c" in type_hint_lower:
                pin_type = PinType.NO_CONNECT

        pins.append(
            PinDef(
                number=pin_num,
                name=pin_name,
                pin_type=pin_type,
                side=detect_pin_side(pin_name, pin_type),
                style=detect_pin_style(pin_name, pin_type),
            )
        )

    return SymbolDef(name=name, pins=pins)


def generate_symbol_sexp(sym: SymbolDef) -> str:
    """Generate KiCad S-expression for symbol."""

    # Group pins by side
    pins_by_side = {side: [] for side in PinSide}
    for pin in sym.pins:
        side = pin.side or PinSide.LEFT
        pins_by_side[side].append(pin)

    # Calculate dimensions
    left_count = len(pins_by_side[PinSide.LEFT])
    right_count = len(pins_by_side[PinSide.RIGHT])
    top_count = len(pins_by_side[PinSide.TOP])
    bottom_count = len(pins_by_side[PinSide.BOTTOM])

    max_side = max(left_count, right_count)
    height = max(max_side * sym.pin_spacing, 5.08)  # At least 200 mil

    max_top_bottom = max(top_count, bottom_count)
    width = max(max_top_bottom * sym.pin_spacing, sym.min_width)

    # Ensure width accommodates the longest pin name
    max_name_len = max((len(p.name) for p in sym.pins), default=0)
    min_name_width = max_name_len * 1.27 + sym.pin_length * 2 + 5.08
    width = max(width, min_name_width)

    # Round to grid
    width = round(width / 2.54) * 2.54
    height = round(height / 2.54) * 2.54

    half_w = width / 2
    half_h = height / 2

    # Generate pins
    pin_lines = []

    # Left side pins (pointing right, angle=0)
    for i, pin in enumerate(pins_by_side[PinSide.LEFT]):
        y = half_h - sym.pin_spacing * (i + 0.5) if left_count > 0 else 0
        y = round(y / 1.27) * 1.27  # Snap to 50 mil grid
        x = -half_w - sym.pin_length
        pin_lines.append(_gen_pin_sexp(pin, x, y, 0, sym.pin_length))

    # Right side pins (pointing left, angle=180)
    for i, pin in enumerate(pins_by_side[PinSide.RIGHT]):
        y = half_h - sym.pin_spacing * (i + 0.5) if right_count > 0 else 0
        y = round(y / 1.27) * 1.27
        x = half_w + sym.pin_length
        pin_lines.append(_gen_pin_sexp(pin, x, y, 180, sym.pin_length))

    # Top pins (pointing down, angle=270)
    for i, pin in enumerate(pins_by_side[PinSide.TOP]):
        x = -half_w + sym.pin_spacing * (i + 0.5) if top_count > 0 else 0
        x = round(x / 1.27) * 1.27
        y = half_h + sym.pin_length
        pin_lines.append(_gen_pin_sexp(pin, x, y, 270, sym.pin_length))

    # Bottom pins (pointing up, angle=90)
    for i, pin in enumerate(pins_by_side[PinSide.BOTTOM]):
        x = -half_w + sym.pin_spacing * (i + 0.5) if bottom_count > 0 else 0
        x = round(x / 1.27) * 1.27
        y = -half_h - sym.pin_length
        pin_lines.append(_gen_pin_sexp(pin, x, y, 90, sym.pin_length))

    pins_sexp = "\n".join(pin_lines)

    # Build complete symbol
    sexp = f'''(kicad_symbol_lib
\t(version 20231120)
\t(generator "create_symbol.py")
\t(generator_version "1.0")
\t(symbol "{sym.name}"
\t\t(exclude_from_sim no)
\t\t(in_bom yes)
\t\t(on_board yes)
\t\t(property "Reference" "{sym.reference}"
\t\t\t(at 0 {half_h + 2.54:.2f} 0)
\t\t\t(effects
\t\t\t\t(font
\t\t\t\t\t(size 1.27 1.27)
\t\t\t\t)
\t\t\t)
\t\t)
\t\t(property "Value" "{sym.value}"
\t\t\t(at 0 {-half_h - 2.54:.2f} 0)
\t\t\t(effects
\t\t\t\t(font
\t\t\t\t\t(size 1.27 1.27)
\t\t\t\t)
\t\t\t)
\t\t)
\t\t(property "Footprint" "{sym.footprint}"
\t\t\t(at 0 {-half_h - 5.08:.2f} 0)
\t\t\t(effects
\t\t\t\t(font
\t\t\t\t\t(size 1.27 1.27)
\t\t\t\t)
\t\t\t\t(hide yes)
\t\t\t)
\t\t)
\t\t(property "Datasheet" "{sym.datasheet}"
\t\t\t(at 0 {-half_h - 7.62:.2f} 0)
\t\t\t(effects
\t\t\t\t(font
\t\t\t\t\t(size 1.27 1.27)
\t\t\t\t)
\t\t\t\t(hide yes)
\t\t\t)
\t\t)
\t\t(property "Description" "{sym.description}"
\t\t\t(at 0 {-half_h - 10.16:.2f} 0)
\t\t\t(effects
\t\t\t\t(font
\t\t\t\t\t(size 1.27 1.27)
\t\t\t\t)
\t\t\t\t(hide yes)
\t\t\t)
\t\t)
\t\t(property "ki_keywords" "{sym.keywords}"
\t\t\t(at 0 0 0)
\t\t\t(effects
\t\t\t\t(font
\t\t\t\t\t(size 1.27 1.27)
\t\t\t\t)
\t\t\t\t(hide yes)
\t\t\t)
\t\t)
\t\t(symbol "{sym.name}_0_1"
\t\t\t(rectangle
\t\t\t\t(start {-half_w:.2f} {half_h:.2f})
\t\t\t\t(end {half_w:.2f} {-half_h:.2f})
\t\t\t\t(stroke
\t\t\t\t\t(width 0.254)
\t\t\t\t\t(type default)
\t\t\t\t)
\t\t\t\t(fill
\t\t\t\t\t(type background)
\t\t\t\t)
\t\t\t)
\t\t)
\t\t(symbol "{sym.name}_1_1"
{pins_sexp}
\t\t)
\t)
)'''

    return sexp


def _gen_pin_sexp(pin: PinDef, x: float, y: float, angle: float, length: float) -> str:
    """Generate S-expression for a single pin."""

    hidden = "\n\t\t\t\t(hide yes)" if pin.hidden else ""

    return f'''\t\t\t(pin {pin.pin_type.value} {pin.style.value}
\t\t\t\t(at {x:.2f} {y:.2f} {angle:.0f})
\t\t\t\t(length {length:.2f})
\t\t\t\t(name "{pin.name}"
\t\t\t\t\t(effects
\t\t\t\t\t\t(font
\t\t\t\t\t\t\t(size 1.27 1.27)
\t\t\t\t\t\t)
\t\t\t\t\t)
\t\t\t\t)
\t\t\t\t(number "{pin.number}"
\t\t\t\t\t(effects
\t\t\t\t\t\t(font
\t\t\t\t\t\t\t(size 1.27 1.27)
\t\t\t\t\t\t)
\t\t\t\t\t){hidden}
\t\t\t\t)
\t\t\t)'''


# Package templates for common IC configurations
PACKAGE_TEMPLATES = {
    "dip8": {
        "pins": 8,
        "layout": "dual",  # Pins on left and right
        "left_pins": [1, 2, 3, 4],
        "right_pins": [8, 7, 6, 5],  # Reversed (standard DIP)
    },
    "dip14": {
        "pins": 14,
        "layout": "dual",
        "left_pins": list(range(1, 8)),
        "right_pins": list(range(14, 7, -1)),
    },
    "dip16": {
        "pins": 16,
        "layout": "dual",
        "left_pins": list(range(1, 9)),
        "right_pins": list(range(16, 8, -1)),
    },
    "soic8": {
        "pins": 8,
        "layout": "dual",
        "left_pins": [1, 2, 3, 4],
        "right_pins": [8, 7, 6, 5],
    },
    "soic14": {
        "pins": 14,
        "layout": "dual",
        "left_pins": list(range(1, 8)),
        "right_pins": list(range(14, 7, -1)),
    },
    "soic16": {
        "pins": 16,
        "layout": "dual",
        "left_pins": list(range(1, 9)),
        "right_pins": list(range(16, 8, -1)),
    },
    "tssop20": {
        "pins": 20,
        "layout": "dual",
        "left_pins": list(range(1, 11)),
        "right_pins": list(range(20, 10, -1)),
    },
    "tssop28": {
        "pins": 28,
        "layout": "dual",
        "left_pins": list(range(1, 15)),
        "right_pins": list(range(28, 14, -1)),
    },
    "qfp32": {
        "pins": 32,
        "layout": "quad",
        "left_pins": list(range(1, 9)),
        "bottom_pins": list(range(9, 17)),
        "right_pins": list(range(17, 25)),
        "top_pins": list(range(25, 33)),
    },
    "qfp48": {
        "pins": 48,
        "layout": "quad",
        "left_pins": list(range(1, 13)),
        "bottom_pins": list(range(13, 25)),
        "right_pins": list(range(25, 37)),
        "top_pins": list(range(37, 49)),
    },
    "qfp64": {
        "pins": 64,
        "layout": "quad",
        "left_pins": list(range(1, 17)),
        "bottom_pins": list(range(17, 33)),
        "right_pins": list(range(33, 49)),
        "top_pins": list(range(49, 65)),
    },
}


def apply_template(sym: SymbolDef, template_name: str):
    """Apply a package template to pin layout."""
    if template_name not in PACKAGE_TEMPLATES:
        raise ValueError(
            f"Unknown template: {template_name}. Available: {list(PACKAGE_TEMPLATES.keys())}"
        )

    template = PACKAGE_TEMPLATES[template_name]

    # Create a map of pin number to pin object
    pin_map = {p.number: p for p in sym.pins}

    # Assign sides based on template
    if template["layout"] == "dual":
        for num in template["left_pins"]:
            if str(num) in pin_map:
                pin_map[str(num)].side = PinSide.LEFT
        for num in template["right_pins"]:
            if str(num) in pin_map:
                pin_map[str(num)].side = PinSide.RIGHT

    elif template["layout"] == "quad":
        for num in template.get("left_pins", []):
            if str(num) in pin_map:
                pin_map[str(num)].side = PinSide.LEFT
        for num in template.get("right_pins", []):
            if str(num) in pin_map:
                pin_map[str(num)].side = PinSide.RIGHT
        for num in template.get("top_pins", []):
            if str(num) in pin_map:
                pin_map[str(num)].side = PinSide.TOP
        for num in template.get("bottom_pins", []):
            if str(num) in pin_map:
                pin_map[str(num)].side = PinSide.BOTTOM


def create_pins_from_template(template_name: str) -> list[PinDef]:
    """Create placeholder pins from a template."""
    if template_name not in PACKAGE_TEMPLATES:
        raise ValueError(f"Unknown template: {template_name}")

    template = PACKAGE_TEMPLATES[template_name]
    pins = []

    for i in range(1, template["pins"] + 1):
        side = PinSide.LEFT  # Default

        if template["layout"] == "dual":
            if i in template["left_pins"]:
                side = PinSide.LEFT
            elif i in template["right_pins"]:
                side = PinSide.RIGHT
        elif template["layout"] == "quad":
            if i in template.get("left_pins", []):
                side = PinSide.LEFT
            elif i in template.get("right_pins", []):
                side = PinSide.RIGHT
            elif i in template.get("top_pins", []):
                side = PinSide.TOP
            elif i in template.get("bottom_pins", []):
                side = PinSide.BOTTOM

        pins.append(
            PinDef(
                number=str(i),
                name=f"P{i}",
                pin_type=PinType.PASSIVE,
                side=side,
            )
        )

    return pins


def print_pin_summary(sym: SymbolDef):
    """Print a summary of pins grouped by type and side."""
    print(f"\nSymbol: {sym.name}")
    print(f"  Reference: {sym.reference}")
    print(f"  Total pins: {len(sym.pins)}")

    # Group by type
    by_type = {}
    for pin in sym.pins:
        by_type.setdefault(pin.pin_type.value, []).append(pin)

    print("\n  Pins by type:")
    for ptype, pins in sorted(by_type.items()):
        pin_list = ", ".join(f"{p.name}({p.number})" for p in pins)
        print(f"    {ptype}: {pin_list}")

    # Group by side
    by_side = {}
    for pin in sym.pins:
        side = pin.side or PinSide.LEFT
        by_side.setdefault(side.value, []).append(pin)

    print("\n  Pins by side:")
    for side, pins in sorted(by_side.items()):
        pin_list = ", ".join(f"{p.name}({p.number})" for p in pins)
        print(f"    {side}: {pin_list}")


def main():
    parser = argparse.ArgumentParser(
        description="Generate KiCad symbol library files",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog="""
Examples:
  # From JSON definition
  %(prog)s --json tpa3116d2.json --output Amplifier.kicad_sym

  # From CSV (number,name,type columns)
  %(prog)s --csv pcm5122.csv --name PCM5122 --output DAC.kicad_sym

  # From template with names added
  %(prog)s --template soic8 --name NE555 --output Timer.kicad_sym

  # Detect pin types from names
  %(prog)s --csv pins.csv --name MyChip --auto-types --output MyLib.kicad_sym
""",
    )

    input_group = parser.add_mutually_exclusive_group(required=True)
    input_group.add_argument("--json", type=Path, help="JSON pin definition file")
    input_group.add_argument("--csv", type=Path, help="CSV pin definition file")
    input_group.add_argument(
        "--template", choices=list(PACKAGE_TEMPLATES.keys()), help="Use package template"
    )
    input_group.add_argument("--interactive", action="store_true", help="Enter pins interactively")
    input_group.add_argument("--text", type=Path, help="Datasheet-style text file")

    parser.add_argument(
        "--name", "-n", required=False, help="Symbol name (required for CSV/template/interactive)"
    )
    parser.add_argument("--output", "-o", type=Path, required=True, help="Output .kicad_sym file")
    parser.add_argument("--reference", "-r", default="U", help="Reference designator (default: U)")
    parser.add_argument("--footprint", "-f", default="", help="Default footprint")
    parser.add_argument("--description", "-d", default="", help="Symbol description")
    parser.add_argument("--datasheet", default="", help="Datasheet URL")
    parser.add_argument("--keywords", "-k", default="", help="Search keywords")
    parser.add_argument(
        "--auto-types", action="store_true", help="Auto-detect pin types from names"
    )
    parser.add_argument(
        "--auto-sides", action="store_true", help="Auto-assign pin sides based on type"
    )
    parser.add_argument(
        "--apply-template",
        choices=list(PACKAGE_TEMPLATES.keys()),
        help="Apply layout from package template",
    )
    parser.add_argument(
        "--summary", action="store_true", help="Print pin summary before generating"
    )
    parser.add_argument(
        "--dry-run", action="store_true", help="Show what would be generated without writing"
    )

    args = parser.parse_args()

    # Check name is provided for non-JSON inputs
    if not args.json and not args.name:
        parser.error("--name is required for CSV, template, text, and interactive modes")

    # Parse input
    if args.json:
        sym = parse_json(args.json)
    elif args.csv:
        sym = parse_csv(args.csv, args.name)
    elif args.template:
        pins = create_pins_from_template(args.template)
        sym = SymbolDef(name=args.name, pins=pins)
    elif args.text:
        text = args.text.read_text()
        sym = parse_datasheet_text(text, args.name)
    elif args.interactive:
        print("Enter pins (number name [type]), one per line. Empty line to finish:")
        lines = []
        while True:
            try:
                line = input()
                if not line.strip():
                    break
                lines.append(line)
            except EOFError:
                break
        text = "\n".join(lines)
        sym = parse_datasheet_text(text, args.name)
    else:
        parser.error("No input source specified")
        return

    # Apply overrides
    if args.reference:
        sym.reference = args.reference
    if args.footprint:
        sym.footprint = args.footprint
    if args.description:
        sym.description = args.description
    if args.datasheet:
        sym.datasheet = args.datasheet
    if args.keywords:
        sym.keywords = args.keywords

    # Auto-detect types if requested
    if args.auto_types:
        for pin in sym.pins:
            if pin.pin_type == PinType.PASSIVE:
                pin.pin_type = detect_pin_type(pin.name)
                pin.style = detect_pin_style(pin.name, pin.pin_type)

    # Auto-assign sides if requested
    if args.auto_sides:
        for pin in sym.pins:
            if pin.side is None:
                pin.side = detect_pin_side(pin.name, pin.pin_type)

    # Apply template layout
    if args.apply_template:
        apply_template(sym, args.apply_template)

    # Print summary if requested
    if args.summary or args.dry_run:
        print_pin_summary(sym)

    # Generate S-expression
    sexp = generate_symbol_sexp(sym)

    if args.dry_run:
        print("\n--- Generated S-expression ---")
        print(sexp[:2000] + "..." if len(sexp) > 2000 else sexp)
        print(f"\n(Total {len(sexp)} characters)")
        return

    # Write output
    ensure_parent_dir(args.output).write_text(sexp)
    print(f"\nWrote {args.output} ({len(sexp)} bytes)")
    print(f"  Symbol: {sym.name}")
    print(f"  Pins: {len(sym.pins)}")


if __name__ == "__main__":
    main()



"""
PCB Blocks - Virtual Components for Hierarchical PCB Layout.

This module extends the circuit block concept to PCB layout, treating
groups of components as "virtual components" with:
- Internal component placement (relative positions)
- Internal routing (pre-routed critical traces)
- External ports (connection points on the block boundary)

This enables a "divide and conquer" approach to PCB layout:
1. Define blocks with internal placement + routing
2. Place blocks on PCB
3. Route inter-block connections (simpler problem)

Usage:
    from kicad_tools.pcb.blocks import PCBBlock, MCUBlock, LDOBlock

    # Create MCU block with bypass caps pre-placed and pre-routed
    mcu = MCUBlock(
        mcu_footprint="QFP-20_4x4mm",
        bypass_caps=["C12", "C13"],
    )

    # Place on PCB at position
    mcu.place(x=100, y=50, rotation=0)

    # Get port positions for inter-block routing
    vdd_port = mcu.port("VDD")
    pa0_port = mcu.port("PA0")
"""

# Re-export geometry types
from ..geometry import Layer, Point, Rectangle

# Re-export placement types
from ..placement import ComponentPlacement, get_footprint_pads

# Re-export primitive types
from ..primitives import Pad, Port, TraceSegment, Via

# Block classes
from .base import PCBBlock
from .led import LEDBlock
from .mcu import MCUBlock
from .oscillator import OscillatorBlock
from .power import LDOBlock


def __getattr__(name: str):
    """Lazy import for layout and exporter to avoid circular imports."""
    if name == "PCBLayout":
        from ..layout import PCBLayout

        return PCBLayout
    if name == "KiCadPCBExporter":
        from ..exporter import KiCadPCBExporter

        return KiCadPCBExporter
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")


__all__ = [
    # Geometry
    "Point",
    "Rectangle",
    "Layer",
    # Primitives
    "Pad",
    "Port",
    "TraceSegment",
    "Via",
    # Placement
    "ComponentPlacement",
    "get_footprint_pads",
    # Layout (lazy import)
    "PCBLayout",
    # Exporter (lazy import)
    "KiCadPCBExporter",
    # Base class
    "PCBBlock",
    # Block types
    "MCUBlock",
    "LDOBlock",
    "OscillatorBlock",
    "LEDBlock",
]



"""
Routing grid for PCB autorouting.

This module provides:
- RoutingGrid: 3D grid for routing with obstacle tracking and congestion awareness
"""

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from kicad_tools.schema.pcb import Zone

from kicad_tools.exceptions import RoutingError

from .layers import Layer, LayerStack
from .primitives import GridCell, Obstacle, Pad, Route, Segment, Via
from .rules import DesignRules


class RoutingGrid:
    """3D grid for routing with obstacle tracking and congestion awareness."""

    def __init__(
        self,
        width: float,
        height: float,
        rules: DesignRules,
        origin_x: float = 0,
        origin_y: float = 0,
        layer_stack: LayerStack | None = None,
    ):
        self.width = width
        self.height = height
        self.rules = rules
        self.origin_x = origin_x
        self.origin_y = origin_y
        self.resolution = rules.grid_resolution

        # Layer stack (default to 2-layer for backward compatibility)
        self.layer_stack = layer_stack or LayerStack.two_layer()
        self.num_layers = self.layer_stack.num_layers

        # Build layer enum to grid index mapping
        self._layer_to_index: dict[int, int] = {}
        self._index_to_layer: dict[int, int] = {}
        for layer_def in self.layer_stack.layers:
            for layer_enum in Layer:
                if layer_enum.kicad_name == layer_def.name:
                    self._layer_to_index[layer_enum.value] = layer_def.index
                    self._index_to_layer[layer_def.index] = layer_enum.value
                    break

        # Grid dimensions
        self.cols = int(width / self.resolution) + 1
        self.rows = int(height / self.resolution) + 1

        # 3D grid: [layer][y][x]
        self.grid: list[list[list[GridCell]]] = [
            [[GridCell(x, y, layer) for x in range(self.cols)] for y in range(self.rows)]
            for layer in range(self.num_layers)
        ]

        # Congestion tracking: coarser grid for density
        self.congestion_size = rules.congestion_grid_size
        self.congestion_cols = max(1, self.cols // self.congestion_size)
        self.congestion_rows = max(1, self.rows // self.congestion_size)

        # Congestion counts: [layer][cy][cx] = number of blocked cells
        self.congestion: list[list[list[int]]] = [
            [[0 for _ in range(self.congestion_cols)] for _ in range(self.congestion_rows)]
            for _ in range(self.num_layers)
        ]

        # Track placed routes for net assignment
        self.routes: list[Route] = []

        # Alias for backward compatibility
        self.layers = self.num_layers

    def layer_to_index(self, layer_enum_value: int) -> int:
        """Map Layer enum value to grid index."""
        if layer_enum_value in self._layer_to_index:
            return self._layer_to_index[layer_enum_value]
        raise RoutingError(
            "Layer value not in stack",
            context={
                "layer_value": layer_enum_value,
                "available": list(self._layer_to_index.keys()),
            },
        )

    def index_to_layer(self, index: int) -> int:
        """Map grid index to Layer enum value."""
        if index in self._index_to_layer:
            return self._index_to_layer[index]
        raise RoutingError(
            "Grid index not in stack",
            context={"index": index, "available": list(self._index_to_layer.keys())},
        )

    def get_routable_indices(self) -> list[int]:
        """Get grid indices of routable signal layers."""
        return self.layer_stack.get_routable_indices()

    def is_plane_layer(self, index: int) -> bool:
        """Check if grid index is a plane layer (no routing)."""
        return self.layer_stack.is_plane_layer(index)

    def _update_congestion(self, gx: int, gy: int, layer: int, delta: int = 1) -> None:
        """Update congestion count for the region containing (gx, gy)."""
        cx = min(gx // self.congestion_size, self.congestion_cols - 1)
        cy = min(gy // self.congestion_size, self.congestion_rows - 1)
        self.congestion[layer][cy][cx] += delta

    def get_congestion(self, gx: int, gy: int, layer: int) -> float:
        """Get congestion level [0, 1] for a grid cell's region."""
        cx = min(gx // self.congestion_size, self.congestion_cols - 1)
        cy = min(gy // self.congestion_size, self.congestion_rows - 1)
        count = self.congestion[layer][cy][cx]
        max_cells = self.congestion_size * self.congestion_size
        return min(1.0, count / max_cells)

    def get_congestion_map(self) -> dict[str, float]:
        """Get congestion statistics for all regions."""
        stats: dict[str, float] = {
            "max_congestion": 0.0,
            "avg_congestion": 0.0,
            "congested_regions": 0,
        }

        total = 0.0
        count = 0
        max_cells = self.congestion_size * self.congestion_size

        for layer in range(self.layers):
            for cy in range(self.congestion_rows):
                for cx in range(self.congestion_cols):
                    density = self.congestion[layer][cy][cx] / max_cells
                    total += density
                    count += 1
                    stats["max_congestion"] = max(stats["max_congestion"], density)
                    if density > self.rules.congestion_threshold:
                        stats["congested_regions"] += 1

        stats["avg_congestion"] = total / count if count > 0 else 0.0
        return stats

    def world_to_grid(self, x: float, y: float) -> tuple[int, int]:
        """Convert world coordinates to grid indices.

        Uses round() instead of int() to avoid floating point precision errors.
        For example, (112.6 - 75.0) / 0.1 = 375.9999999999999 should map to 376,
        but int() would truncate to 375, causing off-by-one grid cell errors.
        """
        gx = round((x - self.origin_x) / self.resolution)
        gy = round((y - self.origin_y) / self.resolution)
        return (max(0, min(gx, self.cols - 1)), max(0, min(gy, self.rows - 1)))

    def grid_to_world(self, gx: int, gy: int) -> tuple[float, float]:
        """Convert grid indices to world coordinates."""
        return (
            self.origin_x + gx * self.resolution,
            self.origin_y + gy * self.resolution,
        )

    def add_obstacle(self, obs: Obstacle) -> None:
        """Mark grid cells as blocked by an obstacle."""
        clearance = obs.clearance + self.rules.trace_clearance

        # Calculate affected grid region
        x1 = obs.x - obs.width / 2 - clearance
        y1 = obs.y - obs.height / 2 - clearance
        x2 = obs.x + obs.width / 2 + clearance
        y2 = obs.y + obs.height / 2 + clearance

        gx1, gy1 = self.world_to_grid(x1, y1)
        gx2, gy2 = self.world_to_grid(x2, y2)

        layer_idx = self.layer_to_index(obs.layer.value)

        for gy in range(gy1, gy2 + 1):
            for gx in range(gx1, gx2 + 1):
                if 0 <= gx < self.cols and 0 <= gy < self.rows:
                    self.grid[layer_idx][gy][gx].blocked = True

    def add_pad(self, pad: Pad) -> None:
        """Add a pad as an obstacle (except for its own net)."""
        # Clearance model: only trace clearance from pad edge
        # The trace itself doesn't need additional margin since we're measuring
        # from pad edge to trace edge, not center to center
        clearance = self.rules.trace_clearance

        if pad.through_hole:
            if pad.width > 0 and pad.height > 0:
                effective_width = pad.width
                effective_height = pad.height
            elif pad.drill > 0:
                effective_width = pad.drill + 0.7
                effective_height = effective_width
            else:
                effective_width = 1.7
                effective_height = 1.7
        else:
            effective_width = pad.width
            effective_height = pad.height

        x1 = pad.x - effective_width / 2 - clearance
        y1 = pad.y - effective_height / 2 - clearance
        x2 = pad.x + effective_width / 2 + clearance
        y2 = pad.y + effective_height / 2 + clearance

        gx1, gy1 = self.world_to_grid(x1, y1)
        gx2, gy2 = self.world_to_grid(x2, y2)

        # PTH pads block all layers, SMD pads block only their layer
        if pad.through_hole:
            layers_to_block = list(range(self.num_layers))
        else:
            layers_to_block = [self.layer_to_index(pad.layer.value)]

        # Get center cell coordinates
        center_gx, center_gy = self.world_to_grid(pad.x, pad.y)

        # Calculate pad metal area bounds (without clearance)
        metal_x1 = pad.x - effective_width / 2
        metal_y1 = pad.y - effective_height / 2
        metal_x2 = pad.x + effective_width / 2
        metal_y2 = pad.y + effective_height / 2
        metal_gx1, metal_gy1 = self.world_to_grid(metal_x1, metal_y1)
        metal_gx2, metal_gy2 = self.world_to_grid(metal_x2, metal_y2)

        for layer_idx in layers_to_block:
            for gy in range(gy1, gy2 + 1):
                for gx in range(gx1, gx2 + 1):
                    if 0 <= gx < self.cols and 0 <= gy < self.rows:
                        cell = self.grid[layer_idx][gy][gx]
                        cell.blocked = True
                        # Mark as pad-blocked so route rip-up won't corrupt it
                        cell.pad_blocked = True
                        cell.original_net = pad.net

                        is_metal_area = (
                            metal_gx1 <= gx <= metal_gx2 and metal_gy1 <= gy <= metal_gy2
                        )

                        if is_metal_area:
                            if cell.net == 0:
                                cell.net = pad.net
                            elif cell.net != pad.net and pad.net != 0:
                                cell.is_obstacle = True
                        else:
                            if pad.net == 0:
                                if cell.net != 0:
                                    cell.is_obstacle = True
                            elif cell.net == 0:
                                cell.net = pad.net
                            elif cell.net != pad.net:
                                cell.is_obstacle = True

            # Always mark the center cell with this pad's net
            if 0 <= center_gx < self.cols and 0 <= center_gy < self.rows:
                center_cell = self.grid[layer_idx][center_gy][center_gx]
                center_cell.net = pad.net
                center_cell.original_net = pad.net

    def add_keepout(
        self,
        x1: float,
        y1: float,
        x2: float,
        y2: float,
        layers: list[Layer] | None = None,
    ) -> None:
        """Add a keepout region."""
        if layers is None:
            layer_indices = self.get_routable_indices()
        else:
            layer_indices = [self.layer_to_index(layer.value) for layer in layers]

        gx1, gy1 = self.world_to_grid(x1, y1)
        gx2, gy2 = self.world_to_grid(x2, y2)

        for layer_idx in layer_indices:
            for gy in range(gy1, gy2 + 1):
                for gx in range(gx1, gx2 + 1):
                    if 0 <= gx < self.cols and 0 <= gy < self.rows:
                        self.grid[layer_idx][gy][gx].blocked = True

    def is_blocked(self, gx: int, gy: int, layer: Layer, net: int = 0) -> bool:
        """Check if a cell is blocked for routing."""
        if not (0 <= gx < self.cols and 0 <= gy < self.rows):
            return True
        layer_idx = self.layer_to_index(layer.value)
        cell = self.grid[layer_idx][gy][gx]
        if cell.blocked:
            return cell.net == 0 or cell.net != net
        return False

    def mark_route(self, route: Route) -> None:
        """Mark a route's cells as used."""
        total_clearance = self.rules.trace_width / 2 + self.rules.trace_clearance
        clearance_cells = int(total_clearance / self.resolution) + 1

        for seg in route.segments:
            self._mark_segment(seg, clearance_cells=clearance_cells)
        for via in route.vias:
            self._mark_via(via)
        self.routes.append(route)

    def _mark_segment(self, seg: Segment, clearance_cells: int = 1) -> None:
        """Mark cells along a segment as blocked (with clearance buffer)."""
        gx1, gy1 = self.world_to_grid(seg.x1, seg.y1)
        gx2, gy2 = self.world_to_grid(seg.x2, seg.y2)

        layer_idx = self.layer_to_index(seg.layer.value)
        marked_cells: set[tuple[int, int]] = set()

        def mark_with_clearance(gx: int, gy: int) -> None:
            for dy in range(-clearance_cells, clearance_cells + 1):
                for dx in range(-clearance_cells, clearance_cells + 1):
                    nx, ny = gx + dx, gy + dy
                    if 0 <= nx < self.cols and 0 <= ny < self.rows:
                        cell = self.grid[layer_idx][ny][nx]
                        if not cell.blocked:
                            # First time blocking - this is a route cell
                            marked_cells.add((nx, ny))
                            cell.net = seg.net
                        # else: cell already blocked (by pad), don't change net
                        cell.blocked = True

        # Simple line marking
        if gx1 == gx2:  # Vertical
            for gy in range(min(gy1, gy2), max(gy1, gy2) + 1):
                mark_with_clearance(gx1, gy)
        elif gy1 == gy2:  # Horizontal
            for gx in range(min(gx1, gx2), max(gx1, gx2) + 1):
                mark_with_clearance(gx, gy1)
        else:  # Diagonal - use Bresenham
            dx = abs(gx2 - gx1)
            dy = abs(gy2 - gy1)
            sx = 1 if gx1 < gx2 else -1
            sy = 1 if gy1 < gy2 else -1
            err = dx - dy
            gx, gy = gx1, gy1
            while True:
                mark_with_clearance(gx, gy)
                if gx == gx2 and gy == gy2:
                    break
                e2 = 2 * err
                if e2 > -dy:
                    err -= dy
                    gx += sx
                if e2 < dx:
                    err += dx
                    gy += sy

        # Update congestion for all newly marked cells
        for nx, ny in marked_cells:
            self._update_congestion(nx, ny, layer_idx)

    def _mark_via(self, via: Via) -> None:
        """Mark cells around a via as blocked on ALL layers (through-hole via)."""
        gx, gy = self.world_to_grid(via.x, via.y)
        radius = int((via.diameter / 2 + self.rules.via_clearance) / self.resolution)

        for layer_idx in range(self.num_layers):
            for dy in range(-radius, radius + 1):
                for dx in range(-radius, radius + 1):
                    nx, ny = gx + dx, gy + dy
                    if 0 <= nx < self.cols and 0 <= ny < self.rows:
                        cell = self.grid[layer_idx][ny][nx]
                        if not cell.blocked:
                            self._update_congestion(nx, ny, layer_idx)
                            cell.net = via.net
                        cell.blocked = True

    def unmark_route(self, route: Route) -> None:
        """Unmark a route's cells (rip-up). Reverses mark_route()."""
        total_clearance = self.rules.trace_width / 2 + self.rules.trace_clearance
        clearance_cells = int(total_clearance / self.resolution) + 1

        for seg in route.segments:
            self._unmark_segment(seg, clearance_cells=clearance_cells)
        for via in route.vias:
            self._unmark_via(via)

        if route in self.routes:
            self.routes.remove(route)

    def _unmark_segment(self, seg: Segment, clearance_cells: int = 1) -> None:
        """Unmark cells along a segment (clear blocked status and net)."""
        gx1, gy1 = self.world_to_grid(seg.x1, seg.y1)
        gx2, gy2 = self.world_to_grid(seg.x2, seg.y2)

        layer_idx = self.layer_to_index(seg.layer.value)

        def unmark_with_clearance(gx: int, gy: int) -> None:
            for dy in range(-clearance_cells, clearance_cells + 1):
                for dx in range(-clearance_cells, clearance_cells + 1):
                    nx, ny = gx + dx, gy + dy
                    if 0 <= nx < self.cols and 0 <= ny < self.rows:
                        cell = self.grid[layer_idx][ny][nx]
                        if cell.pad_blocked:
                            # Don't unblock pad cells, just restore original net
                            cell.net = cell.original_net
                        elif cell.net == seg.net:
                            cell.blocked = False
                            cell.net = 0

        if gx1 == gx2:
            for gy in range(min(gy1, gy2), max(gy1, gy2) + 1):
                unmark_with_clearance(gx1, gy)
        elif gy1 == gy2:
            for gx in range(min(gx1, gx2), max(gx1, gx2) + 1):
                unmark_with_clearance(gx, gy1)
        else:
            dx = abs(gx2 - gx1)
            dy = abs(gy2 - gy1)
            sx = 1 if gx1 < gx2 else -1
            sy = 1 if gy1 < gy2 else -1
            err = dx - dy
            gx, gy = gx1, gy1
            while True:
                unmark_with_clearance(gx, gy)
                if gx == gx2 and gy == gy2:
                    break
                e2 = 2 * err
                if e2 > -dy:
                    err -= dy
                    gx += sx
                if e2 < dx:
                    err += dx
                    gy += sy

    def _unmark_via(self, via: Via) -> None:
        """Unmark cells around a via on ALL layers."""
        gx, gy = self.world_to_grid(via.x, via.y)
        radius = int((via.diameter / 2 + self.rules.via_clearance) / self.resolution)

        for layer_idx in range(self.num_layers):
            for dy in range(-radius, radius + 1):
                for dx in range(-radius, radius + 1):
                    nx, ny = gx + dx, gy + dy
                    if 0 <= nx < self.cols and 0 <= ny < self.rows:
                        cell = self.grid[layer_idx][ny][nx]
                        if cell.pad_blocked:
                            # Don't unblock pad cells, just restore original net
                            cell.net = cell.original_net
                        elif cell.net == via.net:
                            cell.blocked = False
                            cell.net = 0

    # =========================================================================
    # NEGOTIATED CONGESTION ROUTING SUPPORT
    # =========================================================================

    def reset_route_usage(self) -> None:
        """Reset all usage counts (start of new negotiation iteration)."""
        for layer_idx in range(self.layers):
            for gy in range(self.rows):
                for gx in range(self.cols):
                    self.grid[layer_idx][gy][gx].usage_count = 0

    def mark_route_usage(
        self, route: Route, net_cells: dict[int, set] | None = None
    ) -> set[tuple[int, int, int]]:
        """Mark cells used by a route, incrementing usage count."""
        cells_used: set[tuple[int, int, int]] = set()

        for seg in route.segments:
            seg_cells = self._get_segment_cells(seg)
            cells_used.update(seg_cells)

        for via in route.vias:
            via_cells = self._get_via_cells(via)
            cells_used.update(via_cells)

        for gx, gy, layer_idx in cells_used:
            if 0 <= gx < self.cols and 0 <= gy < self.rows:
                self.grid[layer_idx][gy][gx].usage_count += 1

        if net_cells is not None:
            if route.net not in net_cells:
                net_cells[route.net] = set()
            net_cells[route.net].update(cells_used)

        return cells_used

    def unmark_route_usage(self, route: Route, net_cells: dict[int, set] | None = None) -> None:
        """Remove a route's usage (rip-up), decrementing usage count."""
        cells_used: set[tuple[int, int, int]] = set()

        for seg in route.segments:
            seg_cells = self._get_segment_cells(seg)
            cells_used.update(seg_cells)

        for via in route.vias:
            via_cells = self._get_via_cells(via)
            cells_used.update(via_cells)

        for gx, gy, layer_idx in cells_used:
            if 0 <= gx < self.cols and 0 <= gy < self.rows:
                cell = self.grid[layer_idx][gy][gx]
                cell.usage_count = max(0, cell.usage_count - 1)

        if net_cells is not None and route.net in net_cells:
            net_cells[route.net] -= cells_used

    def _get_segment_cells(self, seg: Segment) -> set[tuple[int, int, int]]:
        """Get all grid cells occupied by a segment."""
        cells: set[tuple[int, int, int]] = set()
        gx1, gy1 = self.world_to_grid(seg.x1, seg.y1)
        gx2, gy2 = self.world_to_grid(seg.x2, seg.y2)
        layer_idx = self.layer_to_index(seg.layer.value)

        if gx1 == gx2:
            for gy in range(min(gy1, gy2), max(gy1, gy2) + 1):
                cells.add((gx1, gy, layer_idx))
        elif gy1 == gy2:
            for gx in range(min(gx1, gx2), max(gx1, gx2) + 1):
                cells.add((gx, gy1, layer_idx))
        else:
            dx = abs(gx2 - gx1)
            dy = abs(gy2 - gy1)
            sx = 1 if gx1 < gx2 else -1
            sy = 1 if gy1 < gy2 else -1
            err = dx - dy
            gx, gy = gx1, gy1
            while True:
                cells.add((gx, gy, layer_idx))
                if gx == gx2 and gy == gy2:
                    break
                e2 = 2 * err
                if e2 > -dy:
                    err -= dy
                    gx += sx
                if e2 < dx:
                    err += dx
                    gy += sy
        return cells

    def _get_via_cells(self, via: Via) -> set[tuple[int, int, int]]:
        """Get all grid cells occupied by a via (all layers for through-hole)."""
        cells: set[tuple[int, int, int]] = set()
        gx, gy = self.world_to_grid(via.x, via.y)
        for layer_idx in range(self.num_layers):
            cells.add((gx, gy, layer_idx))
        return cells

    def find_overused_cells(self) -> list[tuple[int, int, int, int]]:
        """Find cells with usage_count > 1 (resource conflicts)."""
        overused = []
        for layer_idx in range(self.layers):
            for gy in range(self.rows):
                for gx in range(self.cols):
                    cell = self.grid[layer_idx][gy][gx]
                    if cell.usage_count > 1:
                        overused.append((gx, gy, layer_idx, cell.usage_count))
        return overused

    def update_history_costs(self, history_increment: float = 1.0) -> None:
        """Increase history cost for overused cells (PathFinder-style)."""
        for layer_idx in range(self.layers):
            for gy in range(self.rows):
                for gx in range(self.cols):
                    cell = self.grid[layer_idx][gy][gx]
                    if cell.usage_count > 1:
                        cell.history_cost += history_increment * (cell.usage_count - 1)

    def get_negotiated_cost(
        self, gx: int, gy: int, layer: int, present_cost_factor: float = 1.0
    ) -> float:
        """Get the negotiated congestion cost for a cell."""
        if not (0 <= gx < self.cols and 0 <= gy < self.rows):
            return float("inf")

        cell = self.grid[layer][gy][gx]

        if cell.is_obstacle:
            return float("inf")

        present_cost = present_cost_factor * cell.usage_count
        history_cost = cell.history_cost

        return present_cost + history_cost

    def get_total_overflow(self) -> int:
        """Get total overflow (sum of usage_count - 1 for overused cells)."""
        overflow = 0
        for layer_idx in range(self.layers):
            for gy in range(self.rows):
                for gx in range(self.cols):
                    usage = self.grid[layer_idx][gy][gx].usage_count
                    if usage > 1:
                        overflow += usage - 1
        return overflow

    # =========================================================================
    # ZONE (COPPER POUR) SUPPORT
    # =========================================================================

    def add_zone_cells(
        self,
        zone: "Zone",
        filled_cells: set[tuple[int, int]],
        layer_index: int,
    ) -> None:
        """Mark grid cells as belonging to a zone.

        Args:
            zone: Zone definition (for net and uuid)
            filled_cells: Set of (gx, gy) grid coordinates to mark
            layer_index: Grid layer index
        """
        from kicad_tools.schema.pcb import Zone as ZoneType  # noqa: F401

        for gx, gy in filled_cells:
            if 0 <= gx < self.cols and 0 <= gy < self.rows:
                cell = self.grid[layer_index][gy][gx]
                cell.is_zone = True
                cell.zone_id = zone.uuid
                cell.net = zone.net_number
                # Zone copper is not an obstacle - routes can pass through same-net zones

    def clear_zones(self, layer_index: int | None = None) -> None:
        """Remove all zone markings from the grid.

        Args:
            layer_index: If specified, only clear this layer. Otherwise clear all.
        """
        layers_to_clear = [layer_index] if layer_index is not None else range(self.num_layers)

        for layer_idx in layers_to_clear:
            for gy in range(self.rows):
                for gx in range(self.cols):
                    cell = self.grid[layer_idx][gy][gx]
                    if cell.is_zone:
                        cell.is_zone = False
                        cell.zone_id = None
                        # Only clear net if it was set by zone (not by a pad)
                        if not cell.is_obstacle and not cell.blocked:
                            cell.net = 0

    def get_zone_cells(self, layer_index: int, zone_id: str | None = None) -> set[tuple[int, int]]:
        """Get all cells belonging to zones on a layer.

        Args:
            layer_index: Grid layer index
            zone_id: If specified, only return cells for this zone

        Returns:
            Set of (gx, gy) coordinates
        """
        cells: set[tuple[int, int]] = set()

        for gy in range(self.rows):
            for gx in range(self.cols):
                cell = self.grid[layer_index][gy][gx]
                if cell.is_zone and (zone_id is None or cell.zone_id == zone_id):
                    cells.add((gx, gy))

        return cells

    def is_zone_cell(self, gx: int, gy: int, layer_index: int) -> bool:
        """Check if a cell is part of a zone.

        Args:
            gx, gy: Grid coordinates
            layer_index: Grid layer index

        Returns:
            True if cell is marked as zone copper
        """
        if not (0 <= gx < self.cols and 0 <= gy < self.rows):
            return False
        return self.grid[layer_index][gy][gx].is_zone

    # =========================================================================
    # BOARD EDGE CLEARANCE SUPPORT
    # =========================================================================

    def add_edge_keepout(
        self,
        edge_segments: list[tuple[tuple[float, float], tuple[float, float]]],
        clearance: float,
    ) -> int:
        """Block cells within clearance distance of board edge segments.

        This prevents routes from being placed too close to the board edge,
        which would violate copper-to-edge clearance DRC rules.

        Args:
            edge_segments: List of (start, end) tuples defining edge line segments.
                          Each segment is ((x1, y1), (x2, y2)) in world coordinates.
            clearance: Edge clearance distance in mm.

        Returns:
            Number of cells blocked.
        """
        if clearance <= 0 or not edge_segments:
            return 0

        blocked_count = 0
        clearance_cells = int(clearance / self.resolution) + 1

        # Get all routable layer indices
        layer_indices = self.get_routable_indices()

        for (x1, y1), (x2, y2) in edge_segments:
            # Mark cells along each edge segment with clearance buffer
            blocked_count += self._mark_edge_segment_keepout(
                x1, y1, x2, y2, clearance_cells, layer_indices
            )

        return blocked_count

    def _mark_edge_segment_keepout(
        self,
        x1: float,
        y1: float,
        x2: float,
        y2: float,
        clearance_cells: int,
        layer_indices: list[int],
    ) -> int:
        """Mark cells within clearance of a single edge segment as blocked.

        Uses Bresenham's algorithm to walk along the segment and blocks all
        cells within the clearance distance on all routable layers.

        Args:
            x1, y1: Start point in world coordinates
            x2, y2: End point in world coordinates
            clearance_cells: Number of grid cells for clearance buffer
            layer_indices: Grid indices of layers to block

        Returns:
            Number of cells blocked.
        """
        gx1, gy1 = self.world_to_grid(x1, y1)
        gx2, gy2 = self.world_to_grid(x2, y2)

        blocked_count = 0
        blocked_cells: set[tuple[int, int]] = set()

        def mark_with_clearance(gx: int, gy: int) -> None:
            """Mark cells within clearance radius of a point."""
            nonlocal blocked_count
            for dy in range(-clearance_cells, clearance_cells + 1):
                for dx in range(-clearance_cells, clearance_cells + 1):
                    nx, ny = gx + dx, gy + dy
                    if (nx, ny) in blocked_cells:
                        continue
                    if 0 <= nx < self.cols and 0 <= ny < self.rows:
                        # Check if within circular clearance (not square)
                        if dx * dx + dy * dy <= clearance_cells * clearance_cells:
                            blocked_cells.add((nx, ny))
                            for layer_idx in layer_indices:
                                cell = self.grid[layer_idx][ny][nx]
                                if not cell.blocked:
                                    cell.blocked = True
                                    cell.is_obstacle = True
                                    blocked_count += 1

        # Walk along the segment using Bresenham's algorithm
        if gx1 == gx2:  # Vertical line
            for gy in range(min(gy1, gy2), max(gy1, gy2) + 1):
                mark_with_clearance(gx1, gy)
        elif gy1 == gy2:  # Horizontal line
            for gx in range(min(gx1, gx2), max(gx1, gx2) + 1):
                mark_with_clearance(gx, gy1)
        else:  # Diagonal - use Bresenham
            dx = abs(gx2 - gx1)
            dy = abs(gy2 - gy1)
            sx = 1 if gx1 < gx2 else -1
            sy = 1 if gy1 < gy2 else -1
            err = dx - dy
            gx, gy = gx1, gy1
            while True:
                mark_with_clearance(gx, gy)
                if gx == gx2 and gy == gy2:
                    break
                e2 = 2 * err
                if e2 > -dy:
                    err -= dy
                    gx += sx
                if e2 < dx:
                    err += dx
                    gy += sy

        return blocked_count
